#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.18 04:00:00                  #
# ================================================== #

from pygpt_net.plugin.base_plugin import BasePlugin


class Plugin(BasePlugin):
    def __init__(self):
        super(Plugin, self).__init__()
        self.id = "self_loop"
        self.name = "Self Loop"
        self.description = "Allows to execute self talk (AI to AI) loop and connect output to input."
        self.iteration = 0
        self.prev_output = None
        self.window = None
        self.order = 9998
        self.use_locale = True
        self.init_options()

    def init_options(self):
        """
        Initialize options
        """
        self.add_option("iterations", "int", 3,
                        "Iterations",
                        "How many iterations to run? 0 = infinite.\nWARNING: setting this to 0 can cause a lot of "
                        "requests and heavy tokens usage!",
                        min=0, max=100, multiplier=1, step=1, slider=True)
        self.add_option("clear_output", "bool", True,
                        "Clear context output",
                        "If enabled, previous context output will be cleared before sending new input.")
        self.add_option("reverse_roles", "bool", True,
                        "Reverse roles between iterations",
                        "If enabled, roles will be reversed between iterations.")

    def setup(self):
        """
        Return available config options

        :return: config options
        """
        return self.options

    def attach(self, window):
        """
        Attach window

        :param window: Window instance
        """
        self.window = window

    def handle(self, event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        """
        name = event.name
        data = event.data
        ctx = event.ctx

        if name == 'ctx.before':
            self.on_ctx_before(ctx)
        elif name == 'ctx.after':
            self.on_ctx_after(ctx)
        elif name == 'ctx.end':
            self.on_ctx_end(ctx)
        elif name == 'user.send':
            self.on_user_send(data['value'])

    def on_user_send(self, text):
        """
        Event: On user send text

        :param text: text
        """
        self.iteration = 0
        self.prev_output = None

    def on_ctx_end(self, ctx):
        """
        Event: On context end

        :param ctx: CtxItem
        """
        iterations = int(self.get_option_value("iterations"))
        if iterations == 0 or self.iteration < iterations:
            self.iteration += 1
            if self.prev_output is not None and self.prev_output != "":
                self.debug(
                    "Plugin: self_loop:on_ctx_end: {}".format(self.prev_output))  # log
                self.window.controller.input.send(self.prev_output)

    def on_ctx_before(self, ctx):
        """
        Event: Before ctx

        :param ctx: CtxItem
        """
        if self.iteration > 0 and self.iteration % 2 != 0 and self.get_option_value("reverse_roles"):
            self.debug("Plugin: self_loop:on_ctx_before [before]: {}".format(ctx.dump()))  # log
            tmp_input_name = ctx.input_name
            tmp_output_name = ctx.output_name
            ctx.input_name = tmp_output_name
            ctx.output_name = tmp_input_name
            self.debug("Plugin: self_loop:on_ctx_before [after]: {}".format(ctx.dump()))  # log

    def on_ctx_after(self, ctx):
        """
        Event: After ctx

        :param ctx: CtxItem
        """
        self.prev_output = ctx.output
        if self.get_option_value("clear_output"):
            self.debug("Plugin: self_loop:on_ctx_after [before]: {}".format(ctx.dump()))  # log
            ctx.output = ""
            self.debug("Plugin: self_loop:on_ctx_after [after]: {}".format(ctx.dump()))  # log



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

import webbrowser

from pygpt_net.utils import trans


class Assistant:
    def __init__(self, window=None):
        """
        Assistants controller

        :param window: Window instance
        """
        self.window = window

    def setup(self):
        """Setup assistants"""
        self.window.app.assistants.load()
        self.update()

    def update(self, update_list=True):
        """
        Update assistants list

        :param update_list: update list
        """
        if update_list:
            self.update_list()
        self.window.controller.assistant_files.update_uploaded()
        self.select_assistant_by_current()

    def update_list(self):
        """Update assistants list"""
        items = self.window.app.assistants.get_all()
        self.window.ui.toolbox.assistants.update(items)

    def assistant_change_locked(self):
        """
        Check if assistant change is locked

        :return: true if locked
        :rtype: bool
        """
        if self.window.controller.input.generating:
            return True
        return False

    def select(self, idx):
        """
        Select assistant on the list

        :param idx: idx on the list
        """
        # check if change is not locked
        if self.assistant_change_locked():
            return

        # mark assistant as selected
        id = self.window.app.assistants.get_by_idx(idx)
        self.select_by_id(id)

    def select_by_id(self, id):
        """
        Select assistant on the list

        :param id: assistant ID
        """
        self.window.app.config.set('assistant', id)

        # update attachments list with list of attachments from assistant
        mode = self.window.app.config.get('mode')
        assistant = self.window.app.assistants.get_by_id(id)
        self.window.controller.attachment.import_from_assistant(mode, assistant)
        self.window.controller.attachment.update()
        self.update(False)

        # update model if exists in assistant
        if assistant is not None:
            model = assistant.model
            if model is not None and model != "":
                if model in self.window.app.models.items:
                    self.window.app.config.set('model', model)
                    self.window.app.config.data['current_model'][mode] = model
                    self.update_assistants()

        self.window.controller.ctx.update_ctx()  # update current ctx info

    def select_assistant_by_current(self):
        """Select assistant by current"""
        assistant_id = self.window.app.config.get('assistant')
        items = self.window.app.assistants.get_all()
        if assistant_id in items:
            idx = list(items.keys()).index(assistant_id)
            current = self.window.ui.models['assistants'].index(idx, 0)
            self.window.ui.nodes['assistants'].setCurrentIndex(current)

    def select_default_assistant(self):
        """Set default assistant"""
        assistant = self.window.app.config.get('assistant')
        if assistant is None or assistant == "":
            mode = self.window.app.config.get('mode')
            if mode == 'assistant':
                self.window.app.config.set('assistant', self.window.app.assistants.get_default_assistant())
                self.update()

    def update_assistants(self):
        """Update assistants"""
        self.select_default_assistant()

    def update_field(self, id, value, assistant_id=None, current=False):
        """
        Update assistant field from editor

        :param id: field id
        :param value: field value
        :param assistant_id: assistant ID
        :param current: if true, updates current assistant
        """
        if assistant_id is not None and assistant_id != "":
            if self.window.app.assistants.has(assistant_id):
                assistant = self.window.app.assistants.get_by_id(assistant_id)
                if id == 'assistant.name':
                    assistant.name = value
                elif id == 'assistant.description':
                    assistant.description = value
                elif id == 'assistant.instructions':
                    assistant.instructions = value
                elif id == 'assistant.model':
                    assistant.model = value

    def init_editor(self, id=None):
        """
        Initialize assistant editor

        :param id: assistant ID
        """
        assistant = self.window.app.assistants.create()
        assistant.model = "gpt-4-1106-preview"  # default model

        # if editing existing assistant
        if id is not None and id != "":
            if self.window.app.assistants.has(id):
                assistant = self.window.app.assistants.get_by_id(id)
        else:
            # default instructions
            assistant.instructions = 'You are a helpful assistant.'

        if assistant.name is None:
            assistant.name = ""
        if assistant.description is None:
            assistant.description = ""
        if assistant.instructions is None:
            assistant.instructions = ""
        if assistant.model is None:
            assistant.model = ""

        self.window.ui.config_option['assistant.id'].setText(id)
        self.config_change('assistant.name', assistant.name, 'assistant.editor')
        self.config_change('assistant.description', assistant.description, 'assistant.editor')
        self.config_change('assistant.instructions', assistant.instructions, 'assistant.editor')
        self.config_change('assistant.model', assistant.model, 'assistant.editor')

        if assistant.has_tool('code_interpreter'):
            self.config_toggle('assistant.tool.code_interpreter', True, 'assistant.editor')
        else:
            self.config_toggle('assistant.tool.code_interpreter', False, 'assistant.editor')

        if assistant.has_tool('retrieval'):
            self.config_toggle('assistant.tool.retrieval', True, 'assistant.editor')
        else:
            self.config_toggle('assistant.tool.retrieval', False, 'assistant.editor')

        # restore functions
        if assistant.has_functions():
            functions = assistant.get_functions()
            values = []
            for function in functions:
                values.append({"name": function['name'], "params": function['params'], "desc": function['desc']})
            self.window.ui.config_option['assistant.tool.function'].items = values
            self.window.ui.config_option['assistant.tool.function'].model.updateData(values)
        else:
            self.window.ui.config_option['assistant.tool.function'].items = []
            self.window.ui.config_option['assistant.tool.function'].model.updateData([])

        # set focus to name field
        self.window.ui.config_option['assistant.name'].setFocus()

    def edit(self, idx=None):
        """
        Open assistant editor

        :param idx: assistant index (row index)
        """
        id = None
        if idx is not None:
            id = self.window.app.assistants.get_by_idx(idx)

        self.init_editor(id)
        self.window.ui.dialogs.open_editor('editor.assistants', idx)

    def save(self):
        """
        Save assistant
        """
        created = False
        id = self.window.ui.config_option['assistant.id'].text()
        name = self.window.ui.config_option['assistant.name'].text()
        model = self.window.ui.config_option['assistant.model'].text()

        # check name
        if name is None or name == "" or model is None or model == "":
            self.window.ui.dialogs.alert(trans('assistant.form.empty.fields'))
            return

        if id is None or id == "" or not self.window.app.assistants.has(id):
            assistant = self.create()  # id created in API
            if assistant is None:
                print("Assistant not created")
                return
            id = assistant.id
            self.window.app.assistants.add(assistant)
            self.window.ui.config_option['assistant.id'].setText(id)
            created = True
        else:
            assistant = self.window.app.assistants.get_by_id(id)

        # assign data from fields to assistant
        self.assign_data(assistant)

        # update in API
        if not created:
            self.assistant_update(assistant)

        # save file
        self.window.app.assistants.save()
        self.update_assistants()
        self.update()

        self.window.ui.dialogs.close('editor.assistants')
        self.window.set_status(trans('status.assistant.saved'))

        # switch to new assistant
        self.select_by_id(id)

    def create(self):
        """
        Create assistant
        """
        assistant = self.window.app.assistants.create()
        self.assign_data(assistant)
        try:
            return self.window.app.gpt_assistants.create(assistant)
        except Exception as e:
            self.window.app.debug.log(e)
            self.window.ui.dialogs.alert(str(e))

    def assistant_update(self, assistant):
        """
        Update assistant
        """
        self.assign_data(assistant)
        try:
            return self.window.app.gpt_assistants.update(assistant)
        except Exception as e:
            self.window.app.debug.log(e)
            self.window.ui.dialogs.alert(str(e))

    def import_assistants(self, force=False):
        """
        Import all remote assistants from API

        :param force: if true, imports without confirmation
        """
        if not force:
            self.window.ui.dialogs.confirm('assistant_import', '',
                                           trans('confirm.assistant.import'))
            return

        try:
            # import assistants
            items = self.window.app.assistants.get_all()
            self.window.app.gpt_assistants.import_assistants(items)
            self.window.app.assistants.items = items
            self.window.app.assistants.save()

            # import uploaded files
            for id in self.window.app.assistants.items:
                assistant = self.window.app.assistants.get_by_id(id)
                self.window.controller.assistant_files.import_files(assistant)
            # status
            self.window.set_status("Imported assistants: " + str(len(items)))
        except Exception as e:
            self.window.app.debug.log(e)
            print("Error importing assistants")
            self.window.ui.dialogs.alert(str(e))
        self.update()

    def assign_data(self, assistant):
        """
        Assign data from fields to assistant

        :param assistant: assistant
        """
        assistant.name = self.window.ui.config_option['assistant.name'].text()
        assistant.model = self.window.ui.config_option['assistant.model'].text()
        assistant.description = self.window.ui.config_option['assistant.description'].text()
        assistant.instructions = self.window.ui.config_option['assistant.instructions'].toPlainText()
        assistant.tools = {
            'code_interpreter': self.window.ui.config_option['assistant.tool.code_interpreter'].box.isChecked(),
            'retrieval': self.window.ui.config_option['assistant.tool.retrieval'].box.isChecked(),
            'function': [],  # functions are assigned separately
        }

        # assign assistant's functions tool
        functions = []
        for function in self.window.ui.config_option['assistant.tool.function'].items:
            name = function['name']
            params = function['params']
            desc = function['desc']
            if name is None or name == "":
                continue
            if params is None or params == "":
                params = '{"type": "object", "properties": {}}'
            if desc is None:
                desc = ""
            functions.append({"name": name, "params": params, "desc": desc})
        if len(functions) > 0:
            assistant.tools['function'] = functions
        else:
            assistant.tools['function'] = []

    def clear(self, force=False):
        """
        Clear assistant data

        :param force: force clear data
        """
        id = self.window.app.config.get('assistant')

        if not force:
            self.window.ui.dialogs.confirm('assistant_clear', '',
                                           trans('confirm.assistant.clear'))
            return

        if id is not None and id != "":
            if self.window.app.assistants.has(id):
                assistant = self.window.app.assistants.get_by_id(id)
                assistant.reset()

        self.window.set_status(trans('status.assistant.cleared'))
        self.update()

    def delete(self, idx=None, force=False):
        """
        Delete assistant

        :param idx: assistant index (row index)
        :param force: force delete without confirmation
        """
        if idx is not None:
            id = self.window.app.assistants.get_by_idx(idx)
            if id is not None and id != "":
                if self.window.app.assistants.has(id):
                    # if exists then show confirmation dialog
                    if not force:
                        self.window.ui.dialogs.confirm('assistant_delete', idx,
                                                       trans('confirm.assistant.delete'))
                        return

                    # clear if this is current assistant
                    if id == self.window.app.config.get('assistant'):
                        self.window.app.config.set('assistant', None)
                        self.window.app.config.set('assistant_thread', None)

                    # delete in API
                    try:
                        self.window.app.gpt_assistants.delete(id)
                    except Exception as e:
                        self.window.ui.dialogs.alert(str(e))

                    self.window.app.assistants.delete(id)
                    self.update()
                    self.window.set_status(trans('status.assistant.deleted'))

    def config_toggle(self, id, value, section=None):
        """
        Toggle checkbox

        :param id: checkbox option id
        :param value: checkbox option value
        :param section: settings section
        """
        assistant_id = self.window.app.config.get('assistant')  # current assistant
        is_current = True
        if section == 'assistant.editor':
            assistant_id = self.window.ui.config_option['assistant.id']  # editing assistant
            is_current = False
        self.update_field(id, value, assistant_id, is_current)

    def config_change(self, id, value, section=None):
        """
        Change input value

        :param id: input option id
        :param value: input option value
        :param section: settings section
        """
        # validate filename
        if id == 'assistant.id':
            self.window.ui.config_option[id].setText(value)

        assistant_id = self.window.app.config.get('assistant')  # current assistant
        is_current = True
        if section == 'assistant.editor':
            assistant_id = self.window.ui.config_option['assistant.id']  # editing assistant
            is_current = False
        self.update_field(id, value, assistant_id, is_current)
        self.window.ui.config_option[id].setText('{}'.format(value))

    def goto_online(self):
        """Opens Assistants page"""
        webbrowser.open('https://platform.openai.com/assistants')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

import os
import sys
import traceback
import logging

from pathlib import Path

from PySide6.QtCore import Qt
from PySide6.QtGui import QStandardItemModel

from pygpt_net.debug.assistants import AssistantsDebug
from pygpt_net.debug.attachments import AttachmentsDebug
from pygpt_net.debug.config import ConfigDebug
from pygpt_net.debug.context import ContextDebug
from pygpt_net.debug.models import ModelsDebug
from pygpt_net.debug.plugins import PluginsDebug
from pygpt_net.debug.presets import PresetsDebug
from pygpt_net.debug.ui import UIDebug
from pygpt_net.config import Config


class CustomQtMsgType:
    QtDebugMsg = logging.DEBUG
    QtInfoMsg = logging.INFO
    QtWarningMsg = logging.WARNING
    QtCriticalMsg = logging.ERROR
    QtFatalMsg = logging.CRITICAL


class Debug:
    DBG_KEY, DBG_VALUE = range(2)

    def __init__(self, window=None):
        """
        Debugger handler

        :param window: Window instance
        """
        self.window = window

        # setup workers
        self.workers = {}
        self.workers['assistants'] = AssistantsDebug(self.window)
        self.workers['attachments'] = AttachmentsDebug(self.window)
        self.workers['config'] = ConfigDebug(self.window)
        self.workers['context'] = ContextDebug(self.window)
        self.workers['models'] = ModelsDebug(self.window)
        self.workers['plugins'] = PluginsDebug(self.window)
        self.workers['presets'] = PresetsDebug(self.window)
        self.workers['ui'] = UIDebug(self.window)

        # prepare debug ids
        self.ids = self.workers.keys()
        self.models = {}
        self.initialized = {}
        self.active = {}
        self.idx = {}
        self.counters = {}

        # prepare debug workers data
        for id in self.ids:
            self.models[id] = self.create_model(self.window)
            self.initialized[id] = False
            self.active[id] = False
            self.idx[id] = 0

    @staticmethod
    def init(level=logging.ERROR):
        """
        Initialize error handler
        """
        if not os.path.exists(os.path.join(Path.home(), '.config', Config.CONFIG_DIR)):
            os.makedirs(os.path.join(Path.home(), '.config', Config.CONFIG_DIR))

        logging.basicConfig(
            level=level,
            format='%(asctime)s - %(levelname)s - %(message)s',
            filename=str(Path(os.path.join(Path.home(), '.config', Config.CONFIG_DIR, 'error.log'))),
            filemode='a'
        )

        def handle_exception(exc_type, value, tb):
            """
            Handle uncaught exception
            """
            if not hasattr(logging, '_is_handling_exception'):
                logging._is_handling_exception = True
                logging.error("Uncaught exception:", exc_info=(exc_type, value, tb))
                traceback.print_exception(exc_type, value, tb)
                del logging._is_handling_exception  # remove flag when done
            else:
                traceback.print_exception(exc_type, value, tb)

        sys.excepthook = handle_exception

    def log(self, error):
        """
        Handle error

        :param error: error object
        """
        # if error is only string then log and print it
        if not isinstance(error, Exception):
            print("Error: {}".format(str(error)))
            data = f"MSG: {error}\n"
            print(data)
            logging.error(data)
            return

        etype, value, tb = sys.exc_info()
        traceback_details = traceback.extract_tb(tb)
        if len(traceback_details) >= 4:
            last_calls = traceback_details[-4:]
        else:
            last_calls = traceback_details
        formatted_traceback = ''.join(traceback.format_list(last_calls))
        data = f"Type: {etype.__name__}, MSG: " \
               f"{value}\n" \
               f"Traceback:\n{formatted_traceback}"

        logging.error(data)
        print("Error: {}".format(str(error)))
        print(data)

    def update(self, all=False):
        """
        Update debug windows

        :param all: update all debug windows
        """
        not_realtime = ['context']
        for id in self.workers:
            if id in self.active and self.active[id]:
                if all or id not in not_realtime:
                    self.workers[id].update()

    def begin(self, id):
        """
        Begin debug data

        :param id: debug id
        """
        self.window.ui.debug[id].setModel(self.models[id])
        if id not in self.counters or self.counters[id] != self.models[id].rowCount():
            self.models[id].removeRows(0, self.models[id].rowCount())
            self.initialized[id] = False
        self.idx[id] = 0

    def end(self, id):
        """
        End debug data

        :param id: debug id
        """
        self.counters[id] = self.idx[id]
        self.initialized[id] = True

    def add(self, id, k, v):
        """
        Append debug entry

        :param id: debug id
        :param k: key
        :param v: value
        """
        if self.initialized[id] is False:
            idx = self.models[id].rowCount()
            self.models[id].insertRow(idx)
            self.models[id].setData(self.models[id].index(idx, self.DBG_KEY), k)
            self.models[id].setData(self.models[id].index(idx, self.DBG_VALUE), v)
        else:
            for idx in range(0, self.models[id].rowCount()):
                if self.models[id].index(idx, self.DBG_KEY).data() == k:
                    self.models[id].setData(self.models[id].index(idx, self.DBG_VALUE), v)
                    self.idx[id] += 1
                    return
        self.idx[id] += 1

    def create_model(self, parent):
        """
        Create list model

        :param parent: parent widget
        :return: model instance
        :rtype: QStandardItemModel
        """
        model = QStandardItemModel(0, 2, parent)
        model.setHeaderData(self.DBG_KEY, Qt.Horizontal, "Key")
        model.setHeaderData(self.DBG_VALUE, Qt.Horizontal, "Value")
        return model



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

import datetime
import json
import os
import uuid

from pygpt_net.provider.notepad.base import BaseProvider
from pygpt_net.item.notepad import NotepadItem


class JsonFileProvider(BaseProvider):
    def __init__(self, window=None):
        super(JsonFileProvider, self).__init__(window)
        self.window = window
        self.id = "json_file"
        self.type = "notepad"
        self.config_file = 'notepad.json'

    def create_id(self):
        """
        Create unique uuid

        :return: uuid
        :rtype: str
        """
        return str(uuid.uuid4())

    def create(self, notepad):
        """
        Create new and return its ID

        :param notepad: NotepadItem
        :return: notepad ID
        :rtype: str
        """
        if notepad.id is None or notepad.id == "":
            notepad.id = self.create_id()
        return notepad.id

    def load_all(self):
        """Load notepads from file"""
        path = os.path.join(self.window.app.config.path, self.config_file)
        items = {}
        try:
            if os.path.exists(path):
                with open(path, 'r', encoding="utf-8") as file:
                    data = json.load(file)
                    if data == "" or data is None:
                        return {}

                    # migrate from old version < 2.0.49
                    if 'items' not in data and 'content' in data:
                        ary = {}
                        for id in data['content']:
                            new_id = int(id)
                            notepad = NotepadItem()
                            notepad.id = new_id
                            notepad.title = "Notepad " + str(new_id)
                            notepad.content = data['content'][id]
                            notepad.created_at = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
                            notepad.updated_at = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
                            ary[new_id] = notepad
                        self.save_all(ary)
                        print("Migrated notepads from old version.")
                        return ary

                    # deserialize
                    if 'items' in data:
                        for id in data['items']:
                            item = data['items'][id]
                            notepad = NotepadItem()
                            self.deserialize(item, notepad)
                            id = notepad.id
                            items[id] = notepad
        except Exception as e:
            self.window.app.debug.log(e)
            items = {}

        return items

    def load(self, id):
        """Load notepad from file"""
        path = os.path.join(self.window.app.config.path, self.config_file)
        try:
            if os.path.exists(path):
                with open(path, 'r', encoding="utf-8") as file:
                    data = json.load(file)
                    if data == "" or data is None or 'items' not in data:
                        return {}

                    # deserialize
                    for item in data['items']:
                        tmp_id = item['id']
                        if tmp_id == id:
                            notepad = NotepadItem()
                            self.deserialize(item, notepad)
                            return notepad
        except Exception as e:
            self.window.app.debug.log(e)

    def save(self, notepad):
        """
        Save notepad to file
        """
        try:
            id = notepad.id
            items = self.load_all()  # load all notepads
            items[id] = notepad  # update only current notepad
            self.save_all(items)

        except Exception as e:
            self.window.app.debug.log(e)
            print("Error while saving notepad: {}".format(str(e)))

    def save_all(self, items):
        """
        Save notepad to file
        """
        try:
            # update notepads
            path = os.path.join(self.window.app.config.path, self.config_file)
            data = {}
            ary = {}

            # serialize
            for id in items:
                notepad = items[id]
                ary[id] = self.serialize(notepad)

            data['__meta__'] = self.window.app.config.append_meta()
            data['items'] = ary
            dump = json.dumps(data, indent=4)
            with open(path, 'w', encoding="utf-8") as f:
                f.write(dump)

        except Exception as e:
            self.window.app.debug.log(e)
            print("Error while saving notepad: {}".format(str(e)))

    def remove(self, id):
        """
        Delete by id

        :param id: id
        """
        items = self.load_all()
        if id in items:
            del items[id]
            self.save_all(items)

    def truncate(self):
        """Delete all"""
        path = os.path.join(self.window.app.config.path, self.config_file)
        data = {'__meta__': self.window.app.config.append_meta(), 'items': {}}
        try:
            dump = json.dumps(data, indent=4)
            with open(path, 'w', encoding="utf-8") as f:
                f.write(dump)
        except Exception as e:
            self.window.app.debug.log(e)

    def patch(self, version):
        """
        Migrate presets to current app version

        :param version: current app version
        :return: true if migrated
        :rtype: bool
        """
        return False

    @staticmethod
    def serialize(notepad):
        """
        Serialize item to dict

        :return: serialized item
        :rtype: dict
        """
        return {
            'id': notepad.id,
            'title': notepad.title,
            'content': notepad.content,
            'created_at': notepad.created_at,  # '2019-01-01T00:00:00
            'updated_at': notepad.updated_at,  # '2019-01-01T00:00:00
        }

    @staticmethod
    def deserialize(data, notepad):
        """
        Deserialize item from dict

        :param data: serialized item
        :param notepad: NotepadItem
        """
        if 'id' in data:
            notepad.id = data['id']
        if 'title' in data:
            notepad.title = data['title']
        if 'content' in data:
            notepad.content = data['content']
        if 'created_at' in data:
            notepad.created_at = data['created_at']
        if 'updated_at' in data:
            notepad.updated_at = data['updated_at']

    def dump(self, item):
        """
        Dump to string

        :param item: item to dump
        :return: dumped item as string (json)
        :rtype: str
        """
        return json.dumps(self.serialize(item))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

import os

from PySide6.QtCore import Qt
from PySide6.QtGui import QPixmap
from PySide6.QtWidgets import QDialog, QLabel, QVBoxLayout, QPushButton, QPlainTextEdit, QHBoxLayout

from pygpt_net.utils import trans


class UpdateDialog(QDialog):
    def __init__(self, window=None):
        """
        Update dialog

        :param window: main window
        """
        super(UpdateDialog, self).__init__(window)
        self.window = window
        self.setParent(window)
        self.setWindowTitle(trans('update.title'))

        self.download = QPushButton(trans('update.download'))
        self.download.setCursor(Qt.PointingHandCursor)
        self.download.clicked.connect(
            lambda: self.window.controller.info.goto_update())

        self.snap = QPushButton(trans('update.snap'))
        self.snap.setCursor(Qt.PointingHandCursor)
        self.snap.clicked.connect(
            lambda: self.window.controller.info.goto_snap())

        self.changelog = QPlainTextEdit()
        self.changelog.setReadOnly(True)
        self.changelog.setMinimumHeight(200)

        logo_label = QLabel()
        path = os.path.abspath(
            os.path.join(self.window.app.config.get_root_path(), 'data', 'logo.png'))
        pixmap = QPixmap(path)
        logo_label.setPixmap(pixmap)

        buttons = QHBoxLayout()
        buttons.addWidget(self.download)
        buttons.addWidget(self.snap)

        self.layout = QVBoxLayout()
        self.message = QLabel("")
        self.info = QLabel(trans("update.info"))
        self.info.setWordWrap(True)
        self.info.setAlignment(Qt.AlignCenter)
        self.info.setStyleSheet(self.window.controller.theme.get_style('text_bold'))
        self.info.setStyleSheet("font-weight: bold; font-size: 12px; margin: 20px 0px 20px 0px;")
        self.info.setMaximumHeight(60)
        self.layout.addWidget(logo_label)
        self.layout.addWidget(self.info)
        self.layout.addWidget(self.message)
        self.layout.addWidget(self.changelog, 1)
        self.layout.addLayout(buttons)
        self.layout.addStretch()
        self.setLayout(self.layout)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

from pygpt_net.utils import trans

from PySide6.QtCore import QCoreApplication, QThread


class BasePlugin:
    def __init__(self):
        self.id = ""
        self.name = ""
        self.type = []  # audio.input, audio.output
        self.description = ""
        self.urls = {}
        self.options = {}
        self.initial_options = {}
        self.window = None
        self.parent = None
        self.enabled = False
        self.use_locale = False
        self.is_async = False
        self.order = 0

    def setup(self):
        """
        Return available config options

        :return: config options
        :rtype: dict
        """
        return self.options

    def add_option(self,
                   name,  # option name (ID, key)
                   type,  # option type (text, textarea, bool, int, float, dict)
                   value=None,  # option value
                   label="",  # option label
                   description="",  # option description
                   tooltip=None,  # option tooltip
                   min=None,  # option min value
                   max=None,  # option max value
                   multiplier=1,  # option float value multiplier (for sliders)
                   step=1,  # option step value (for sliders)
                   slider=False,  # option slider (True/False)
                   keys=None,  # option keys (for dict type)
                   advanced=False,  # option is advanced (True/False)
                   secret=False,  # option is secret (True/False)
                   persist=False,  # option is persistent on reset to defaults (True/False)
                   urls=None):  # option URLs (API keys, docs, etc.)
        """
        Add plugin configuration option

        :param name: Option name (ID, key)
        :param type: Option type (text, textarea, bool, int, float, dict)
        :param value: Option value
        :param label: Option label
        :param description: Option description
        :param tooltip: Option tooltip
        :param min: Option min value
        :param max: Option max value
        :param multiplier: Option float value multiplier
        :param step: Option step value (for slider)
        :param slider: Option slider (True/False)
        :param keys: Option keys (for dict type)
        :param advanced: Option advanced (True/False)
        :param secret: Option secret (True/False)
        :param persist: Option persist (True/False)
        :param urls: Option URLs
        """
        if tooltip is None:
            tooltip = description

        option = {
            "type": type,
            "value": value,
            "label": label,
            "description": description,
            "tooltip": tooltip,
            "min": min,
            "max": max,
            "multiplier": multiplier,
            "step": step,
            "slider": slider,
            "keys": keys,
            "advanced": advanced,
            "secret": secret,
            "persist": persist,
            "urls": urls,
        }
        self.options[name] = option

    def has_option(self, name):
        """
        Check if option exists

        :param name: option name
        :return: true if exists
        :rtype: bool
        """
        return name in self.options

    def get_option(self, name):
        """
        Return option

        :param name: option name
        :return: option
        :rtype: dict
        """
        if self.has_option(name):
            return self.options[name]

    def get_option_value(self, name):
        """
        Return option value

        :param name: option name
        :return: option value
        :rtype: any
        """
        if self.has_option(name):
            return self.options[name]["value"]

    def attach(self, window):
        """
        Attach window to plugin

        :param window: Window instance
        """
        self.window = window

    def handle(self, event, *args, **kwargs):
        """
        Handle event

        :param event: event name
        :param args: arguments
        :param kwargs: keyword arguments
        """
        return

    def trans(self, text=None):
        """
        Translate text using plugin domain
        :param text: text to translate
        :return: translated text
        :rtype: str
        """
        if text is None:
            return ""
        domain = 'plugin.{}'.format(self.id)
        return trans(text, False, domain)

    def debug(self, data):
        """
        Send thread-safe debug message to logger window

        :param data: data to send
        """
        if QCoreApplication.instance().thread() == QThread.currentThread():
            self.window.log(data, True)
        else:
            self.window.log(data, False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

import os
import threading
import time

from PySide6.QtCore import QObject, Signal, Slot
import speech_recognition as sr
from openai import OpenAI
import audioop

from pygpt_net.plugin.base_plugin import BasePlugin
from pygpt_net.utils import trans


class Plugin(BasePlugin):
    def __init__(self):
        super(Plugin, self).__init__()
        self.id = "audio_openai_whisper"
        self.name = "Audio Input (OpenAI Whisper)"
        self.type = ['audio.input']
        self.description = "Enables speech recognition using OpenAI Whisper API"
        self.input_text = None
        self.window = None
        self.speech_enabled = False
        self.thread_started = False
        self.thread = None
        self.listening = False
        self.waiting = False
        self.stop = False
        self.magic_word_detected = False
        self.is_first_adjust = True
        self.empty_phrases = ['Thank you for watching']  # phrases to ignore (fix for empty phrases)
        self.order = 1
        self.use_locale = True
        self.init_options()

    def init_options(self):
        """
        Initialize options
        """
        self.add_option("model", "text", "whisper-1",
                        "Model",
                        "Specify model, default: whisper-1")
        self.add_option("timeout", "int", 2,
                        "Timeout",
                        "Speech recognition timeout. Default: 2", min=0, max=30, slider=True,
                        tooltip="Timeout, default: 2")
        self.add_option("phrase_length", "int", 4,
                        "Phrase max length",
                        "Speech recognition phrase length. Default: 4", min=0, max=30, slider=True,
                        tooltip="Phrase max length, default: 4")
        self.add_option("min_energy", "float", 1.3,
                        "Min. energy",
                        "Minimum threshold multiplier above the noise level to begin recording; 1 = disabled. "
                        "Default: 1.3",
                        min=1, max=50, slider=True,
                        tooltip="Min. energy, default: 1.3, 1 = disabled, adjust for your microphone", multiplier=10)
        self.add_option("adjust_noise", "bool", True,
                        "Adjust ambient noise",
                        "Adjust for ambient noise. Default: True")
        self.add_option("continuous_listen", "bool", False,
                        "Continuous listening",
                        "EXPERIMENTAL: continuous listening - do not stop listening after a single input.\n"
                        "Warning: This feature may lead to unexpected results and requires fine-tuning with the rest "
                        "of the options!")
        self.add_option("auto_send", "bool", True,
                        "Auto send",
                        "Automatically send input when voice is detected. Default: True")
        self.add_option("wait_response", "bool", True,
                        "Wait for response",
                        "Wait for a response before listening for the next input. Default: True")
        self.add_option("magic_word", "bool", False,
                        "Magic word",
                        "Activate listening only after the magic word is provided, like 'Hey GPT' or 'OK GPT'. "
                        "Default: False")
        self.add_option("magic_word_reset", "bool", True,
                        "Reset Magic word",
                        "Reset the magic word status after it is received "
                        "(the magic word will need to be provided again). Default: True")
        self.add_option("magic_words", "text", "OK, Okay, Hey GPT, OK GPT",
                        "Magic words",
                        "Specify magic words for 'Magic word' option: if received this word then start listening, "
                        "put words separated by coma, "
                        "Magic word option must be enabled, examples: \"Hey GPT, OK GPT\"")
        self.add_option("magic_word_timeout", "int", 1,
                        "Magic word timeout",
                        "Magic word recognition timeout. Default: 1", min=0, max=30, slider=True,
                        tooltip="Timeout, default: 1")
        self.add_option("magic_word_phrase_length", "int", 2,
                        "Magic word phrase max length",
                        "Magic word phrase length. Default: 2", min=0, max=30, slider=True,
                        tooltip="Phrase length, default: 2")
        self.add_option("prefix_words", "text", "",
                        "Prefix words",
                        "Specify prefix words: if defined, only phrases starting with these words will be transmitted, "
                        "and the remainder will be ignored. Separate the words with a comma., eg. 'OK, Okay, GPT'. "
                        "Leave empty to disable")
        self.add_option("stop_words", "text", "stop, exit, quit, end, finish, close, terminate, kill, halt, abort",
                        "Stop words",
                        "Specify stop words: if any of these words are received, then stop listening. "
                        "Separate the words with a comma, or leave it empty to disable the feature, "
                        "default: stop, exit, quit, end, finish, close, terminate, kill, "
                        "halt, abort")

        # advanced options
        self.add_option("recognition_energy_threshold", "int", 300,
                        "energy_threshold",
                        "Represents the energy level threshold for sounds. Default: 300", min=0, max=10000,
                        slider=True, advanced=True)
        self.add_option("recognition_dynamic_energy_threshold", "bool", True,
                        "dynamic_energy_threshold",
                        "Represents whether the energy level threshold "
                        "for sounds should be automatically adjusted based on the currently "
                        "ambient noise level while listening. Default: True", advanced=True)
        self.add_option("recognition_dynamic_energy_adjustment_damping", "float", 0.15,
                        "dynamic_energy_adjustment_damping",
                        "Represents approximately the fraction of the current energy threshold that "
                        "is retained after one second of dynamic threshold adjustment. Default: 0.15", min=0, max=100,
                        slider=True, multiplier=100, advanced=True)
        self.add_option("recognition_pause_threshold", "float", 0.8,
                        "pause_threshold",
                        "Represents the minimum length of silence (in seconds) that will "
                        "register as the end of a phrase. \nDefault: 0.8",
                        min=0, max=100, slider=True, multiplier=10, advanced=True)
        self.add_option("recognition_adjust_for_ambient_noise_duration", "float", 1,
                        "adjust_for_ambient_noise: duration",
                        "The duration parameter is the maximum number of seconds that it will "
                        "dynamically adjust the threshold for before returning. Default: 1", min=0, max=100,
                        slider=True, multiplier=10, advanced=True)

    def setup(self):
        """
        Return available config options

        :return: config options
        """
        return self.options

    def setup_ui(self):
        """
        Setup UI
        """
        pass

    def get_stop_words(self):
        """
        Return stop words

        :return: stop words
        :rtype: list
        """
        words_str = self.get_option_value('stop_words')
        words = []
        if words_str is not None and len(words_str) > 0 and words_str.strip() != ' ':
            words = words_str.split(',')
            # remove white-spaces
            words = [x.strip() for x in words]
        return words

    def get_magic_words(self):
        """
        Return magic words

        :return: stop words
        :rtype: list
        """
        words_str = self.get_option_value('magic_words')
        words = []
        if words_str is not None and len(words_str) > 0 and words_str.strip() != ' ':
            words = words_str.split(',')
            # remove white-spaces
            words = [x.strip() for x in words]
        return words

    def get_prefix_words(self):
        """
        Return prefix words

        :return: prefix words
        :rtype: list
        """
        words_str = self.get_option_value('prefix_words')
        words = []
        if words_str is not None and len(words_str) > 0 and words_str.strip() != ' ':
            words = words_str.split(',')
            # remove white-spaces
            words = [x.strip() for x in words]
        return words

    def toggle_speech(self, state):
        """
        Toggle speech

        :param state: state to set
        """
        self.speech_enabled = state
        self.window.ui.plugin_addon['audio.input'].btn_toggle.setChecked(state)

        # Start thread if not started
        if state:
            self.listening = True
            self.stop = False
            self.handle_thread()
        else:
            self.listening = False
            self.set_status('')

    def attach(self, window):
        """
        Attach window

        :param window: Window instance
        """
        self.window = window

    def handle(self, event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        """
        name = event.name
        data = event.data
        ctx = event.ctx

        if name == 'input.before':
            self.on_input_before(data['value'])
        elif name == 'ctx.begin':
            self.on_ctx_begin(ctx)
        elif name == 'ctx.end':
            self.on_ctx_end(ctx)
        elif name == 'enable':
            if data['value'] == self.id:
                self.on_enable()
        elif name == 'disable':
            if data['value'] == self.id:
                self.on_disable()
        elif name == 'audio.input.toggle':
            self.toggle_speech(data['value'])
        elif name == 'audio.input.stop':
            self.on_stop()

    def on_ctx_begin(self, ctx):
        """
        Event: On new context begin

        :param ctx: CtxItem
        """
        self.waiting = True

    def on_ctx_end(self, ctx):
        """
        Event: On context end

        :param ctx: CtxItem
        """
        self.waiting = False

    def on_enable(self):
        """Event: On plugin enable"""
        self.speech_enabled = True
        self.handle_thread()

    def on_disable(self):
        """Event: On plugin disable"""
        self.speech_enabled = False
        self.listening = False
        self.stop = True
        self.window.ui.plugin_addon['audio.input'].btn_toggle.setChecked(False)
        self.set_status('')

    def on_stop(self):
        """Event: On input stop"""
        self.stop = True
        self.listening = False
        self.speech_enabled = False
        self.set_status('')

    def on_input_before(self, text):
        """
        Event: Before input

        :param text: text
        """
        self.input_text = text

    def destroy(self):
        """
        Destroy thread
        """
        self.waiting = True
        self.listening = False
        self.thread_started = False
        self.set_status('')

    def handle_thread(self):
        """
        Handle listener thread
        """
        if self.thread_started:
            return

        listener = AudioInputThread(plugin=self)
        listener.finished.connect(self.handle_input)
        listener.destroyed.connect(self.handle_destroy)
        listener.started.connect(self.handle_started)
        listener.stopped.connect(self.handle_stop)
        listener.status_signal.connect(self.handle_status)
        listener.error_signal.connect(self.handle_error)

        self.thread = threading.Thread(target=listener.run)
        self.thread.start()
        self.thread_started = True

    def can_listen(self):
        """
        Check if can listen

        :return: true if can listen
        :rtype: bool
        """
        state = True
        if self.get_option_value('wait_response') and self.window.controller.input.generating:
            state = False
        if self.window.controller.input.locked:
            state = False
        return state

    def set_status(self, status):
        """
        Set status

        :param status: status
        """
        self.window.ui.plugin_addon['audio.input'].set_status(status)

    @Slot(str)
    def handle_input(self, text):
        """
        Insert text to input and send

        :param text: text
        """
        if text is None or text.strip() == '':
            return

        if not self.can_listen():
            return

        # check prefix words
        prefix_words = self.get_prefix_words()
        if len(prefix_words) > 0:
            for word in prefix_words:
                check_text = text.lower().strip()
                check_word = word.lower().strip()
                if not check_text.startswith(check_word):
                    self.window.set_status(trans('audio.speak.ignoring'))
                    self.set_status(trans('audio.speak.ignoring'))
                    return

        # previous magic word detected state
        magic_prev_detected = self.magic_word_detected  # save magic word detected state before checking for magic word

        # check for magic word
        is_magic_word = False
        if self.get_option_value('magic_word'):
            for word in self.get_magic_words():
                # prepare magic word
                check_word = word.lower().replace('.', '')
                check_text = text.lower()
                if check_text.startswith(check_word):
                    is_magic_word = True
                    self.set_status(trans('audio.magic_word.detected'))
                    self.window.set_status(trans('audio.magic_word.detected'))
                    break

        # if magic word enabled
        if self.get_option_value('magic_word'):
            if not is_magic_word:
                if self.get_option_value('magic_word_reset'):
                    self.magic_word_detected = False
            else:
                self.magic_word_detected = True

            # if not previously detected, then abort now
            if not magic_prev_detected:
                if not is_magic_word:
                    self.window.set_status(trans('audio.magic_word.invalid'))
                self.window.ui.nodes['input'].setText(text)
                return
            else:
                self.window.set_status("")

        # update input text
        self.window.ui.nodes['input'].setText(text)

        # send text
        if self.get_option_value('auto_send'):
            self.set_status('...')
            self.window.set_status(trans('audio.speak.sending'))
            self.window.controller.input.send(text)
            self.set_status('')

    @Slot(object)
    def handle_status(self, data):
        """
        Handle thread debug log
        :param data
        """
        self.set_status(str(data))

    @Slot(object)
    def handle_error(self, error):
        """
        Handle thread error
        :param error
        """
        self.window.app.debug.log(error)

    @Slot()
    def handle_destroy(self):
        """
        Insert text to input and send
        """
        self.thread_started = False
        self.set_status('')

    @Slot()
    def handle_started(self):
        """
        Handle listening started
        """
        pass
        # print("Whisper is listening...")

    @Slot()
    def handle_stop(self):
        """
        Stop listening
        """
        self.thread_started = False
        self.listening = False
        self.window.set_status("")
        self.set_status('')
        # print("Whisper stopped listening...")
        self.toggle_speech(False)


class AudioInputThread(QObject):
    # setup signals
    status_signal = Signal(object)
    error_signal = Signal(object)
    finished = Signal(object)
    destroyed = Signal()
    started = Signal()
    stopped = Signal()

    def __init__(self, plugin=None):
        """
        Audio input listener thread
        """
        super().__init__()
        self.plugin = plugin

    def run(self):
        try:
            if not self.plugin.listening:
                return

            client = OpenAI(
                api_key=self.plugin.window.app.config.get('api_key'),
                organization=self.plugin.window.app.config.get('organization_key'),
            )

            print("Starting audio listener....")
            path = os.path.join(self.plugin.window.app.config.path, 'input.wav')

            self.started.emit()
            self.status_signal.emit('')

            with sr.Microphone() as source:
                while self.plugin.listening and not self.plugin.window.is_closing:
                    self.status_signal.emit('')

                    if self.plugin.stop:
                        self.plugin.stop = False
                        self.plugin.listening = False
                        self.status_signal.emit('Stop.')
                        self.stopped.emit()
                        break

                    if not self.plugin.can_listen():
                        time.sleep(0.5)
                        continue

                    try:
                        recognizer = sr.Recognizer()

                        # set recognizer options
                        recognizer.energy_threshold = self.plugin.get_option_value('recognition_energy_threshold')
                        recognizer.dynamic_energy_threshold = \
                            self.plugin.get_option_value('recognition_dynamic_energy_threshold')
                        recognizer.dynamic_energy_adjustment_damping = \
                            self.plugin.get_option_value('recognition_dynamic_energy_adjustment_damping')
                        recognizer.dynamic_energy_adjustment_ratio = \
                            self.plugin.get_option_value('recognition_dynamic_energy_adjustment_ratio')
                        recognizer.pause_threshold = self.plugin.get_option_value('recognition_pause_threshold')
                        adjust_duration = self.plugin.get_option_value('recognition_adjust_for_ambient_noise_duration')

                        # adjust for ambient noise
                        if self.plugin.get_option_value('adjust_noise'):
                            recognizer.adjust_for_ambient_noise(source, duration=adjust_duration)
                            self.plugin.is_first_adjust = False

                        timeout = self.plugin.get_option_value('timeout')
                        phrase_length = self.plugin.get_option_value('phrase_length')

                        # check for magic word, if no magic word detected, then set to magic word timeout and length
                        if self.plugin.get_option_value('magic_word'):
                            if not self.plugin.magic_word_detected:
                                timeout = self.plugin.get_option_value('magic_word_timeout')
                                phrase_length = self.plugin.get_option_value('magic_word_phrase_length')

                        # set begin status
                        if self.plugin.can_listen():
                            if self.plugin.get_option_value('magic_word'):
                                if self.plugin.magic_word_detected:
                                    self.status_signal.emit(trans('audio.speak.now'))
                                else:
                                    self.status_signal.emit(trans('audio.magic_word.please'))
                            else:
                                self.status_signal.emit(trans('audio.speak.now'))

                        min_energy = self.plugin.get_option_value('min_energy')
                        ambient_noise_energy = min_energy * recognizer.energy_threshold

                        if timeout > 0 and phrase_length > 0:
                            audio_data = recognizer.listen(source, timeout, phrase_length)
                        elif timeout > 0:
                            audio_data = recognizer.listen(source, timeout)
                        else:
                            audio_data = recognizer.listen(source)

                        if not self.plugin.can_listen():
                            continue

                        # transcript audio
                        raw_data = audio_data.get_wav_data()
                        is_stop_word = False

                        if raw_data:
                            # check RMS / energy
                            rms = audioop.rms(raw_data, 2)
                            if min_energy > 0:
                                self.status_signal.emit("{}: {} / {} (x{})".
                                                              format(trans('audio.speak.energy'),
                                                                     rms, int(ambient_noise_energy), min_energy))
                            if rms < ambient_noise_energy:
                                continue

                            # save audio file
                            with open(path, "wb") as audio_file:
                                audio_file.write(raw_data)

                            # transcribe
                            with open(path, "rb") as audio_file:
                                self.status_signal.emit(trans('audio.speak.wait'))
                                transcript = client.audio.transcriptions.create(
                                    model=self.plugin.get_option_value('model'),
                                    file=audio_file,
                                    response_format="text"
                                )
                                # handle transcript
                                if transcript is not None and transcript.strip() != '':
                                    # fix if empty phrase
                                    is_empty_phrase = False
                                    transcript_check = transcript.strip().lower()
                                    for phrase in self.plugin.empty_phrases:
                                        phrase_check = phrase.strip().lower()
                                        if phrase_check in transcript_check:
                                            is_empty_phrase = True
                                            break

                                    if is_empty_phrase:
                                        continue

                                    if self.plugin.can_listen():
                                        self.finished.emit(transcript)

                                    # stop listening if not continuous mode or stop word detected
                                    stop_words = self.plugin.get_stop_words()

                                    if len(stop_words) > 0:
                                        is_stop_word = transcript.replace('.', '').strip().lower() in stop_words

                        if not self.plugin.get_option_value('continuous_listen') or is_stop_word:
                            self.plugin.listening = False
                            self.stopped.emit()
                            self.status_signal.emit('')  # clear status
                            break
                    except Exception as e:
                        print("Speech recognition error: {}".format(str(e)))

            self.destroyed.emit()

        except Exception as e:
            self.error_signal.emit(e)
            self.destroyed.emit()
            print("Audio input thread error: {}".format(str(e)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QTextEdit


class OptionTextarea(QTextEdit):
    def __init__(self, window=None, id=None, autoupdate=False, section=None):
        """
        Settings input

        :param window: main window
        :param id: option id
        :param autoupdate: auto update
        :param section: settings section
        """
        super(OptionTextarea, self).__init__(window)
        self.window = window
        self.id = id
        self.section = section
        self.autoupdate = autoupdate
        self.update_ui = True

    def keyPressEvent(self, event):
        """
        Key press event

        :param event: key event
        """
        super(OptionTextarea, self).keyPressEvent(event)
        if not self.autoupdate:
            return
        self.window.controller.settings.change(self.id, self.toPlainText(), self.section)
        if self.update_ui:
            self.window.controller.ui.update()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.23 22:00:00                  #
# ================================================== #

import os

from PySide6.QtCore import Qt
from PySide6.QtGui import QFontDatabase
from PySide6.QtWidgets import QSplitter

from pygpt_net.ui.dialogs import Dialogs
from pygpt_net.ui.layout.chat.main import ChatMain
from pygpt_net.ui.layout.ctx.main import CtxMain
from pygpt_net.ui.layout.toolbox.main import ToolboxMain
from pygpt_net.ui.menu import Menu


class UI:
    def __init__(self, window=None):
        """
        UI (main)

        :param window: Window instance
        """
        self.window = window

        # prepare
        self.config_option = {}
        self.debug = {}
        self.dialog = {}
        self.editor = {}
        self.groups = {}
        self.menu = {}
        self.models = {}
        self.nodes = {}
        self.notepad = {}
        self.parts = {}
        self.paths = {}
        self.plugin_addon = {}
        self.plugin_data = {}
        self.plugin_option = {}
        self.splitters = {}
        self.tabs = {}

        # builders
        self.chat = ChatMain(window)
        self.contexts = CtxMain(window)
        self.dialogs = Dialogs(window)
        self.menus = Menu(window)
        self.toolbox = ToolboxMain(window)

    def init(self):
        """Setup UI"""
        # load font
        self.setup_font()

        # ctx, chat and toolbox
        self.parts['ctx'] = self.contexts.setup()
        self.parts['chat'] = self.chat.setup()
        self.parts['toolbox'] = self.toolbox.setup()

        # set width
        self.parts['ctx'].setMinimumWidth(200)

        # horizontal splitter
        self.splitters['main'] = QSplitter(Qt.Horizontal)
        self.splitters['main'].addWidget(self.parts['ctx'])  # contexts
        self.splitters['main'].addWidget(self.parts['chat'])  # chat box
        self.splitters['main'].addWidget(self.parts['toolbox'])  # toolbox
        self.splitters['main'].setSizes([1, 8, 1])

        # menus
        self.menus.setup()

        # dialogs
        self.dialogs.setup()

        # set central widget
        self.window.setCentralWidget(self.window.ui.splitters['main'])

    def setup_font(self):
        """Setup UI font"""
        path = os.path.join(self.window.app.config.get_root_path(), 'data', 'fonts', 'Lato', 'Lato-Regular.ttf')
        font_id = QFontDatabase.addApplicationFont(path)
        if font_id == -1:
            print("Error loading font file {}".format(path))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

import datetime
import json
import os

from pygpt_net.provider.ctx.base import BaseProvider
from pygpt_net.item.ctx import CtxItem, CtxMeta


class JsonFileProvider(BaseProvider):
    def __init__(self, window=None):
        super(JsonFileProvider, self).__init__(window)
        self.window = window
        self.id = "json_file"
        self.type = "ctx"

    def install(self):
        """
        Install provider data
        """
        context_dir = os.path.join(self.window.app.config.path, 'context')
        if not os.path.exists(context_dir):
            os.mkdir(context_dir)

    def create_id(self):
        """
        Create unique ctx ID

        Format: YYYYMMDDHHMMSS.MICROSECONDS.json

        :return: generated ID
        :rtype: str
        """
        return datetime.datetime.now().strftime("%Y%m%d%H%M%S.%f")

    def create(self, meta):
        """
        Create new ctx and return its ID

        :param meta: CtxMeta
        :return: ctx ID
        """
        if meta.id is None or meta.id == "":
            meta.id = self.create_id()
        return meta.id

    def get_meta(self):
        """Load ctx metadata from file"""
        contexts = {}
        path = os.path.join(self.window.app.config.path, 'context.json')
        try:
            if os.path.exists(path):
                with open(path, 'r', encoding="utf-8") as file:
                    data = json.load(file)
                    if data == "" or data is None or 'items' not in data:
                        return contexts
                    contexts = self.parse_meta(data['items'])
        except Exception as e:
            self.window.app.debug.log(e)
            contexts = {}

        return contexts

    def load(self, id):
        """
        Load ctx data from json file

        :param id: context id
        :return: context items (list of CtxItem)
        :rtype: list
        """
        data = []
        path = os.path.join(self.window.app.config.path, 'context', id + '.json')
        if os.path.exists(path):
            try:
                with open(path, 'r', encoding="utf-8") as file:
                    data = self.parse_data(json.load(file))
                    if data == "" or data is None:
                        return []
            except Exception as e:
                self.window.app.debug.log(e)
                print("Error while loading context: {}".format(id))
                data = []
        return data

    def save(self, id, meta, items):
        """
        Dump ctx to json file

        :param id: ctx id
        :param meta: ctx meta (CtxMeta)
        :param items: ctx items (list of CtxItem)
        """
        try:
            # update current ctx items
            items_path = os.path.join(self.window.app.config.path, 'context', id + '.json')
            serialized_items = []
            for item in items:
                serialized_items.append(self.serialize_item(item))
            dump = json.dumps(serialized_items, indent=4)

            # save items
            with open(items_path, 'w', encoding="utf-8") as f:
                f.write(dump)

            # update ctx meta (index)
            index_path = os.path.join(self.window.app.config.path, 'context.json')
            metas = self.get_meta()
            metas[id] = meta
            data = {}
            serialized_meta = {}
            for n in metas:
                serialized_meta[n] = self.serialize_meta(metas[n])

            data['items'] = serialized_meta
            data['__meta__'] = self.window.app.config.append_meta()
            dump = json.dumps(data, indent=4)

            # save index
            with open(index_path, 'w', encoding="utf-8") as f:
                f.write(dump)

        except Exception as e:
            self.window.app.debug.log(e)
            print("Error while dumping context: {}".format(id))

    def remove(self, id):
        """
        Delete ctx by id

        :param id: ctx id
        """
        # delete ctx meta (index)
        index_path = os.path.join(self.window.app.config.path, 'context.json')
        metas = self.get_meta()
        if id in metas:
            del metas[id]
        data = {}
        serialized_meta = {}
        for n in metas:
            serialized_meta[n] = self.serialize_meta(metas[n])
        data['items'] = serialized_meta
        data['__meta__'] = self.window.app.config.append_meta()
        dump = json.dumps(data, indent=4)

        # save index
        with open(index_path, 'w', encoding="utf-8") as f:
            f.write(dump)

        # delete items
        path = os.path.join(self.window.app.config.path, 'context', id + '.json')
        if os.path.exists(path):
            try:
                os.remove(path)
            except Exception as e:
                self.window.app.debug.log(e)

    def truncate(self):
        """Delete all ctx"""
        # delete all ctx files
        contexts = self.get_meta()
        for id in contexts:
            path = os.path.join(self.window.app.config.path, 'context', id + '.json')
            if os.path.exists(path):
                try:
                    os.remove(path)
                except Exception as e:
                    self.window.app.debug.log(e)

        # truncate ctx index
        index_path = os.path.join(self.window.app.config.path, 'context.json')
        data = {}
        data['items'] = {}
        data['__meta__'] = self.window.app.config.append_meta()
        try:
            dump = json.dumps(data, indent=4)
            with open(index_path, 'w', encoding="utf-8") as f:
                f.write(dump)
        except Exception as e:
            self.window.app.debug.log(e)

    def patch(self, version):
        """
        Migrate presets to current app version

        :param version: current app version
        :return: true if migrated
        :rtype: bool
        """
        return False

    @staticmethod
    def serialize_meta(meta):
        """
        Serialize CtxMeta to dict

        :param meta: CtxMeta
        :return: serialized CtxMeta dict
        :rtype: dict
        """
        return {
            'id': meta.id,
            'name': meta.name,
            'date': meta.date,
            'mode': meta.mode,
            'last_mode': meta.last_mode,
            'thread': meta.thread,
            'assistant': meta.assistant,
            'preset': meta.preset,
            'run': meta.run,
            'status': meta.status,
            'initialized': meta.initialized,
        }

    @staticmethod
    def deserialize_meta(data, meta):
        """
        Deserialize CtxMeta from dict

        :param data: dict
        :param meta: CtxMeta
        """
        if 'id' in data:
            meta.id = data['id']
        if 'name' in data:
            meta.name = data['name']
        if 'date' in data:
            meta.date = data['date']
        if 'mode' in data:
            meta.mode = data['mode']
        if 'last_mode' in data:
            meta.last_mode = data['last_mode']
        if 'thread' in data:
            meta.thread = data['thread']
        if 'assistant' in data:
            meta.assistant = data['assistant']
        if 'preset' in data:
            meta.preset = data['preset']
        if 'run' in data:
            meta.run = data['run']
        if 'status' in data:
            meta.status = data['status']
        if 'initialized' in data:
            meta.initialized = data['initialized']
        return meta

    @staticmethod
    def serialize_item(ctx):
        """
        Serialize CtxItem to dict

        :param ctx: CtxItem
        :return: serialized item dict
        :rtype: dict
        """
        return {
            'input': ctx.input,
            'output': ctx.output,
            'mode': ctx.mode,
            'thread': ctx.thread,
            'msg_id': ctx.msg_id,
            'run_id': ctx.run_id,
            'input_name': ctx.input_name,
            'output_name': ctx.output_name,
            'input_tokens': ctx.input_tokens,
            'output_tokens': ctx.output_tokens,
            'total_tokens': ctx.total_tokens,
            'input_timestamp': ctx.input_timestamp,
            'output_timestamp': ctx.output_timestamp
        }

    @staticmethod
    def deserialize_item(data, ctx):
        """
        Deserialize CtxItem from dict

        :param data: data dict
        :param ctx: CtxItem
        """
        if 'input' in data:
            ctx.input = data['input']
        if 'output' in data:
            ctx.output = data['output']
        if 'mode' in data:
            ctx.mode = data['mode']
        if 'thread' in data:
            ctx.thread = data['thread']
        if 'msg_id' in data:
            ctx.msg_id = data['msg_id']
        if 'run_id' in data:
            ctx.run_id = data['run_id']
        if 'input_name' in data:
            ctx.input_name = data['input_name']
        if 'output_name' in data:
            ctx.output_name = data['output_name']
        if 'input_tokens' in data:
            ctx.input_tokens = data['input_tokens']
        if 'output_tokens' in data:
            ctx.output_tokens = data['output_tokens']
        if 'total_tokens' in data:
            ctx.total_tokens = data['total_tokens']
        if 'input_timestamp' in data:
            ctx.input_timestamp = data['input_timestamp']
        if 'output_timestamp' in data:
            ctx.output_timestamp = data['output_timestamp']

    def dump(self, ctx):
        """
        Dump ctx to string

        :param ctx: CtxItem
        :return: dumped item as string (json)
        :rtype: str
        """
        return json.dumps(self.serialize_item(ctx))

    def parse_data(self, data):
        """
        Parse ctx data from json to objects

        :param data: ctx items data
        :return: ctx items (deserialized) as objects list
        :rtype: list
        """
        items = []
        for item in data:
            ctx = CtxItem()
            self.deserialize_item(item, ctx)
            items.append(ctx)
        return items

    def parse_meta(self, data):
        """
        Parse ctx data from json to objects

        :param data: ctx items data
        :return: ctx items (deserialized) dict
        :rtype: dict
        """
        items = {}
        for k in data:
            meta = CtxMeta()
            self.deserialize_meta(data[k], meta)
            items[meta.id] = meta
        return items



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

from pygpt_net.utils import trans


class Ctx:
    def __init__(self, window=None):
        """
        Context controller

        :param window: Window instance
        """
        self.window = window

        # modes allowed for switch from (key: from, data: to)
        self.allowed_modes = {
            'chat': ['chat', 'completion', 'img', 'langchain', 'vision', 'assistant'],
            'completion': ['chat', 'completion', 'img', 'langchain', 'vision', 'assistant'],
            'img': ['img'],
            'langchain': ['chat', 'completion', 'img', 'langchain', 'vision', 'assistant'],
            'vision': ['chat', 'completion', 'img', 'langchain', 'vision', 'assistant'],
            'assistant': ['assistant'],
        }

    def setup(self):
        """Setup ctx"""
        # load ctx list
        self.window.app.ctx.load_meta()

        # if no context yet then create one
        if len(self.window.app.ctx.meta) == 0:
            self.new()
        else:
            # get last ctx from config
            id = self.window.app.config.get('ctx')
            if id is not None and id in self.window.app.ctx.meta:
                self.window.app.ctx.current = id
            else:
                # if no ctx then get first ctx
                self.window.app.ctx.current = self.window.app.ctx.get_first()

        # load selected ctx
        self.load(self.window.app.ctx.current)

    def update(self, reload=True):
        """
        Update ctx list

        :param reload: reload ctx list items
        """
        # reload ctx list items
        if reload:
            self.reload()
            self.select_ctx_by_current()  # select on list

        # update all
        self.window.controller.ui.update()
        self.window.app.debug.update(True)

        # append ctx and thread id (assistants API) to config
        id = self.window.app.ctx.current
        if id is not None:
            self.window.app.config.set('ctx', id)
            self.window.app.config.set('assistant_thread', self.window.app.ctx.thread)
            self.window.app.config.save()

    def focus_chat(self):
        """Focus chat"""
        # set tab index to 0:
        self.window.ui.tabs['output'].setCurrentIndex(0)

    def select(self, id):
        """
        Select ctx

        :param id: context id
        """
        self.window.app.ctx.current = id
        self.load(id)
        self.focus_chat()

    def select_by_idx(self, idx):
        """
        Select ctx by index

        :param idx: context index
        """
        # lock if generating response is in progress
        if self.context_change_locked():
            return

        id = self.window.app.ctx.get_id_by_idx(idx)
        self.select(id)

    def select_ctx_by_current(self):
        """Select ctx by current"""
        id = self.window.app.ctx.current
        meta = self.window.app.ctx.get_meta()
        if id in meta:
            idx = self.window.app.ctx.get_idx_by_id(id)
            current = self.window.ui.models['ctx.list'].index(idx, 0)
            self.window.ui.nodes['ctx.list'].unlocked = True  # tmp allow change if locked (enable)
            self.window.ui.nodes['ctx.list'].setCurrentIndex(current)
            self.window.ui.nodes['ctx.list'].unlocked = False  # tmp allow change if locked (disable)

    def new(self, force=False):
        """
        Create new ctx

        :param force: force context creation
        """
        # lock if generating response is in progress
        if not force and self.context_change_locked():
            return

        self.window.app.ctx.new()
        self.window.app.config.set('assistant_thread', None)  # reset assistant thread id
        self.update()
        self.window.controller.output.clear()

        if not force:  # only if real click on new context button
            self.window.controller.input.unlock_input()

        # update context label
        mode = self.window.app.ctx.mode
        assistant_id = None
        if mode == 'assistant':
            assistant_id = self.window.app.config.get('assistant')
        self.update_ctx_label(mode, assistant_id)
        self.focus_chat()

    def reload(self):
        """Reload current ctx list"""
        meta = self.window.app.ctx.get_meta()
        self.window.ui.contexts.ctx_list.update('ctx.list', meta)

    def refresh(self):
        """Refresh context"""
        self.load(self.window.app.ctx.current)

    def load(self, ctx):
        """
        Load ctx data

        :param ctx: context name (id)
        """
        # select ctx
        self.window.app.ctx.select(ctx)

        # get current settings stored in ctx
        thread = self.window.app.ctx.thread
        mode = self.window.app.ctx.mode
        assistant_id = self.window.app.ctx.assistant
        preset = self.window.app.ctx.preset

        # restore thread from ctx
        self.window.app.config.set('assistant_thread', thread)

        # clear before output and append ctx to output
        self.window.controller.output.clear()
        self.window.controller.output.append_context()

        # switch mode to ctx mode
        if mode is not None:
            self.window.controller.model.set_mode(mode)  # preset reset here

            # switch preset to ctx preset
            if preset is not None:
                self.window.controller.model.set_preset(mode, preset)
                self.window.controller.model.update_presets()  # update presets only

            # if ctx mode == assistant then switch assistant to ctx assistant
            if mode == 'assistant':
                # if assistant defined then select it
                if assistant_id is not None:
                    self.window.controller.assistant.select_by_id(assistant_id)
                else:
                    # if empty ctx assistant then get assistant from current selected
                    assistant_id = self.window.app.config.get('assistant')

        # reload ctx list and select current ctx on list
        self.update()

        # update current ctx label
        self.update_ctx_label(mode, assistant_id)

    def update_ctx(self):
        """Update current ctx mode if allowed"""
        mode = self.window.app.config.get('mode')

        id = None
        # update ctx mode only if current ctx is allowed for this mode
        if self.is_allowed_for_mode(mode, False):  # do not check assistant match
            self.window.app.ctx.update()

            # update current context label
            if mode == 'assistant':
                if self.window.app.ctx.assistant is not None:
                    # get assistant id from ctx if defined in ctx
                    id = self.window.app.ctx.assistant
                else:
                    # or get assistant id from current selected assistant
                    id = self.window.app.config.get('assistant')

        # update ctx label
        self.update_ctx_label(mode, id)

    def update_ctx_label_by_current(self):
        """
        Update ctx label from current ctx
        """
        mode = self.window.app.ctx.mode

        # if no ctx mode then use current mode
        if mode is None:
            mode = self.window.app.config.get('mode')

        label = trans('mode.' + mode)

        # append assistant name to ctx name label
        if mode == 'assistant':
            id = self.window.app.ctx.assistant
            assistant = self.window.app.assistants.get_by_id(id)
            if assistant is not None:
                # get ctx assistant
                label += ' (' + assistant.name + ')'
            else:
                # get current assistant
                id = self.window.app.config.get('assistant')
                assistant = self.window.app.assistants.get_by_id(id)
                if assistant is not None:
                    label += ' (' + assistant.name + ')'

        # update ctx label
        self.window.controller.ui.update_ctx_label(label)

    def update_ctx_label(self, mode, assistant_id=None):
        """
        Update ctx label

        :param mode: Mode
        :param assistant_id: Assistant id
        """
        if mode is None:
            return
        label = trans('mode.' + mode)
        if mode == 'assistant' and assistant_id is not None:
            assistant = self.window.app.assistants.get_by_id(assistant_id)
            if assistant is not None:
                label += ' (' + assistant.name + ')'

        # update ctx label
        self.window.controller.ui.update_ctx_label(label)

    def delete(self, idx, force=False):
        """
        Delete ctx by idx

        :param idx: context idx
        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm('ctx_delete', idx, trans('ctx.delete.confirm'))
            return

        id = self.window.app.ctx.get_id_by_idx(idx)
        self.window.app.ctx.remove(id)

        # reset current if current ctx deleted
        if self.window.app.ctx.current == id:
            self.window.app.ctx.current = None
            self.window.controller.output.clear()
        self.update()

    def delete_history(self, force=False):
        """
        Delete all ctx / truncate

        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm('ctx_delete_all', '', trans('ctx.delete.all.confirm'))
            return

        # truncate ctx
        self.window.app.ctx.truncate()
        self.update()

    def rename(self, idx):
        """
        Ctx name rename (shows dialog)

        :param idx: context idx
        """
        id = self.window.app.ctx.get_id_by_idx(idx)
        meta = self.window.app.ctx.get_meta_by_id(id)
        self.window.ui.dialog['rename'].id = 'ctx'
        self.window.ui.dialog['rename'].input.setText(meta.name)
        self.window.ui.dialog['rename'].current = id
        self.window.ui.dialog['rename'].show()
        self.update()

    def update_name(self, id, name):
        """
        Update ctx name

        :param id: context id
        :param name: context name
        """
        if id not in self.window.app.ctx.meta:
            return
        self.window.app.ctx.meta[id].name = name
        self.window.app.ctx.set_initialized()
        self.window.app.ctx.save(id)
        self.window.ui.dialog['rename'].close()
        self.update()

    def dismiss_rename(self):
        """Dismiss rename dialog"""
        self.window.ui.dialog['rename'].close()

    def add(self, ctx):
        """
        Add ctx item (CtxItem object)

        :param ctx: CtxItem
        """
        self.window.app.ctx.add(ctx)
        self.update()

    def handle_allowed(self, mode):
        """
        Check if ctx is allowed for this mode, if not then switch to new context

        :param mode: mode name
        :return: True if allowed
        :rtype: bool
        """
        if not self.is_allowed_for_mode(mode):
            self.new(True)  # force new context
            return False
        return True

    def is_allowed_for_mode(self, mode, check_assistant=True):
        """
        Check if ctx is allowed for this mode

        :param mode: mode name
        :param check_assistant: True if check also current assistant
        :return: True if allowed for mode
        :rtype: bool
        """
        # always allow if lock_modes is disabled
        if not self.window.app.config.get('lock_modes'):
            return True

        if self.window.app.ctx.is_empty():
            return True

        # always allow if no ctx
        id = self.window.app.config.get('ctx')
        if id is None or id == '' or not self.window.app.ctx.has(id):
            return True

        meta = self.window.app.ctx.get_meta_by_id(id)

        # always allow if no last mode
        if meta.last_mode is None:
            return True

        # get last used mode from ctx meta
        prev_mode = meta.last_mode
        if prev_mode not in self.allowed_modes[mode]:
            # exception for assistant (if assistant exists in ctx then allow)
            if mode == 'assistant':
                if meta.assistant is not None:
                    # if the same assistant then allow
                    if meta.assistant == self.window.app.config.get('assistant'):
                        return True
                else:
                    return True  # if no assistant in ctx then allow
            # if other mode, then always disallow
            return False

        # check if the same assistant
        if mode == 'assistant' and check_assistant:
            # allow if no assistant yet in ctx
            if meta.assistant is None:
                return True
            # disallow if different assistant
            if meta.assistant != self.window.app.config.get('assistant'):
                return False
        return True

    def selection_change(self):
        """
        Select ctx on list change
        """
        # TODO: implement this
        # idx = self.window.ui.nodes['ctx.list'].currentIndex().row()
        # self.select(idx)
        self.window.ui.nodes['ctx.list'].lockSelection()

    def context_change_locked(self):
        """
        Check if ctx change is locked

        :return: True if locked
        :rtype: bool
        """
        if self.window.controller.input.generating:
            return True
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #


class BaseProvider:
    def __init__(self, window=None):
        self.window = window
        self.id = ""
        self.type = "assistant"

    def install(self):
        pass

    def patch(self, version):
        pass

    def create(self, meta):
        pass

    def load(self):
        pass

    def save(self, items):
        pass

    def remove(self, id):
        pass

    def truncate(self):
        pass

    def dump(self, ctx):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

import datetime

from pygpt_net.item.notepad import NotepadItem
from pygpt_net.provider.notepad.json_file import JsonFileProvider


class Notepad:
    def __init__(self, window=None):
        """
        Notepad

        :param window: Window instance
        """
        self.window = window
        self.providers = {}
        self.provider = "json_file"
        self.items = {}

        # register data providers
        self.add_provider(JsonFileProvider())  # json file provider

    def add_provider(self, provider):
        """
        Add data provider

        :param provider: data provider instance
        """
        self.providers[provider.id] = provider
        self.providers[provider.id].window = self.window

    def install(self):
        """Install provider data"""
        if self.provider in self.providers:
            try:
                self.providers[self.provider].install()
            except Exception as e:
                self.window.app.debug.log(e)

    def patch(self, app_version):
        """Patch provider data"""
        if self.provider in self.providers:
            try:
                self.providers[self.provider].patch(app_version)
            except Exception as e:
                self.window.app.debug.log(e)

    def get_by_id(self, id):
        """
        Get notepad by id

        :param id: notepad id
        :return: notepad instance
        :rtype: NotepadItem
        """
        if id in self.items:
            return self.items[id]
        return None

    def get_all(self):
        """
        Get all notepads

        :return: notepads dict
        :rtype: dict
        """
        return self.items

    def build(self):
        """
        Build notepad

        :param id: notepad id
        :return: NotepadItem instance
        :rtype: NotepadItem
        """
        item = NotepadItem()
        return item

    def add(self, notepad):
        """
        Add notepad

        :param notepad: NotepadItem instance
        :return: True if success
        :rtype: bool
        """
        if self.provider in self.providers:
            try:
                id = self.providers[self.provider].create(notepad)
                notepad.id = id
                self.items[id] = notepad
                self.save(id)
                return True
            except Exception as e:
                self.window.app.debug.log(e)

    def update(self, notepad):
        """
        Update and save notepad

        :param notepad: NotepadItem instance
        :return: True if success
        :rtype: bool
        """
        if notepad.id not in self.items:
            return False

        notepad.updated_at = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        self.items[notepad.id] = notepad
        self.save(notepad.id)
        return True

    def load(self, id):
        """
        Load notepad by id

        :param id: notepad id
        """
        if self.provider in self.providers:
            try:
                self.items[id] = self.providers[self.provider].load(id)
            except Exception as e:
                self.window.app.debug.log(e)

    def load_all(self):
        """Load all notepads"""
        if self.provider in self.providers:
            try:
                self.items = self.providers[self.provider].load_all()
            except Exception as e:
                self.window.app.debug.log(e)

    def save(self, id):
        """
        Save notepad by id

        :param id: notepad id
        :return: True if saved, False if not
        :rtype: bool
        """
        if id not in self.items:
            return False

        if self.provider in self.providers:
            try:
                self.providers[self.provider].save(self.items[id])
                return True
            except Exception as e:
                self.window.app.debug.log(e)
        return False

    def save_all(self):
        """
        Save all notepads
        """
        if self.provider in self.providers:
            try:
                self.providers[self.provider].save_all(self.items)
            except Exception as e:
                self.window.app.debug.log(e)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QCheckBox, QHBoxLayout, QWidget


class OptionCheckbox(QWidget):
    def __init__(self, window=None, id=None, title=None, value=False, section=None):
        """
        Settings checkbox

        :param window: main window
        :param id: option id
        :param title: option title
        :param value: current value
        :param section: settings section
        """
        super(OptionCheckbox, self).__init__(window)
        self.window = window
        self.id = id
        self.title = title
        self.value = value
        self.section = section

        self.box = QCheckBox(title, self.window)
        self.box.setChecked(value)
        self.box.stateChanged.connect(
            lambda: self.window.controller.settings.toggle(self.id, self.box.isChecked(), self.section))

        self.layout = QHBoxLayout()
        self.layout.addWidget(self.box)

        self.setLayout(self.layout)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

import datetime
import os
import requests
from openai import OpenAI


class Image:
    DIRNAME = "img"

    def __init__(self, window=None):
        """
        DALL-E Wrapper

        :param window: Window instance
        """
        self.window = window

    def init(self):
        """Initialize OpenAI API key"""
        pass

    def install(self):
        """Install provider data"""
        img_dir = os.path.join(self.window.app.config.path, self.DIRNAME)
        if not os.path.exists(img_dir):
            os.mkdir(img_dir)

    def get_prompt(self, allow_custom=True):
        """
        Return image generate prompt command

        :param allow_custom: allow custom prompt
        :return: system command for generate image prompt
        """
        cmd = '''
        1. Apply these rules if the request is related to image generation or image description; otherwise, return the user's prompt as is.
        2. Translate any non-English image prompts accurately into English.
        3. Start from "A photograph of..." or "An image of...", etc. DO NOT use asking, like "Please generate...", "I want to see...", etc.
        4. Use as many details as possible to describe the image.
        5. If the user only wants to talk, then return the user's prompt as is (AND ONLY their prompt, without adding any text to it).
        '''
        # get custom prompt from config if exists
        if allow_custom:
            if self.window.app.config.has('img_prompt'):
                prompt = self.window.app.config.get('img_prompt')
                if prompt is not None and prompt != '':
                    cmd = prompt
        return cmd

    def get_client(self):
        """
        Return OpenAI client

        :return: OpenAI client
        """
        return OpenAI(
            api_key=self.window.app.config.get('api_key'),
            organization=self.window.app.config.get('organization_key'),
        )

    def generate(self, prompt, model="dall-e-3", num=1):
        """
        Call DALL-E API

        :param prompt: prompt
        :param model: model name
        :param num: number of variants
        :return: images paths list
        :rtype: list
        """
        if not self.window.app.config.get('img_raw'):
            system_cmd = self.get_prompt()
            max_tokens = 200
            temperature = 1.0
            try:
                # call GPT for generate best image generate prompt
                response = self.window.app.gpt.quick_call(prompt, system_cmd, False, max_tokens,
                                                          self.window.app.config.get('img_prompt_model'), temperature)
                if response is not None and response != "":
                    prompt = response
            except Exception as e:
                self.window.app.debug.log(e)
                print("Image prompt generate by model error: " + str(e))

        print("Generating image from: '{}'".format(prompt))
        client = self.get_client()
        response = client.images.generate(
            model=model,
            prompt=prompt,
            n=num,
            size=self.window.app.config.get('img_resolution'),
        )

        # generate and download images
        paths = []
        for i in range(num):
            if i >= len(response.data):
                break
            url = response.data[i].url
            res = requests.get(url)

            # generate filename
            name = self.make_safe_filename(prompt) + "-" + datetime.date.today().strftime(
                "%Y-%m-%d") + "_" + datetime.datetime.now().strftime("%H-%M-%S") + "-" + str(i + 1) + ".png"
            path = os.path.join(self.window.app.config.path, self.DIRNAME, name)

            print("Downloading... [" + str(i + 1) + " of " + str(num) + "] to: " + path)
            # save image
            if self.save_image(path, res.content):
                paths.append(path)

        return paths, prompt

    def save_image(self, path, image):
        """
        Save image to file

        :param path: path to save
        :param image: image data
        :return: True if success
        """
        try:
            with open(path, 'wb') as file:
                file.write(image)
            return True
        except Exception as e:
            self.window.app.debug.log(e)
            print("Image save error: " + str(e))
            return False

    def make_safe_filename(self, name):
        """
        Make safe filename

        :param name: filename to make safe
        :return: safe filename
        :rtype: str
        """

        def safe_char(c):
            if c.isalnum():
                return c
            else:
                return "_"

        return "".join(safe_char(c) for c in name).rstrip("_")[:30]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

from PySide6.QtGui import QStandardItemModel, Qt
from PySide6.QtWidgets import QVBoxLayout, QLabel, QHBoxLayout, QPushButton, QWidget

from pygpt_net.ui.widget.lists.assistant import AssistantList
from pygpt_net.utils import trans


class Assistants:
    def __init__(self, window=None):
        """
        Toolbox UI

        :param window: Window instance
        """
        self.window = window
        self.id = 'assistants'

    def setup(self):
        """
        Setup assistants

        :return: QWidget
        :rtype: QWidget
        """
        layout = self.setup_assistants()

        self.window.ui.nodes['assistants.widget'] = QWidget()
        self.window.ui.nodes['assistants.widget'].setLayout(layout)
        self.window.ui.nodes['assistants.widget'].setMinimumHeight(150)

        return self.window.ui.nodes['assistants.widget']

    def setup_assistants(self):
        """
        Setup list of assistants

        :return: QVBoxLayout
        :rtype: QVBoxLayout
        """
        # new
        self.window.ui.nodes['assistants.new'] = QPushButton(trans('assistant.new'))
        self.window.ui.nodes['assistants.new'].clicked.connect(
            lambda: self.window.controller.assistant.edit())

        # import
        self.window.ui.nodes['assistants.import'] = QPushButton(trans('assistant.import'))
        self.window.ui.nodes['assistants.import'].clicked.connect(
            lambda: self.window.controller.assistant.import_assistants())

        # label
        self.window.ui.nodes['assistants.label'] = QLabel(trans("toolbox.assistants.label"))
        self.window.ui.nodes['assistants.label'].setStyleSheet(self.window.controller.theme.get_style('text_bold'))

        # header
        header = QHBoxLayout()
        header.addWidget(self.window.ui.nodes['assistants.label'])
        header.addWidget(self.window.ui.nodes['assistants.import'], alignment=Qt.AlignRight)
        header.addWidget(self.window.ui.nodes['assistants.new'], alignment=Qt.AlignRight)
        header.setContentsMargins(0, 0, 0, 0)
        header_widget = QWidget()
        header_widget.setLayout(header)

        # list
        self.window.ui.nodes[self.id] = AssistantList(self.window, self.id)
        self.window.ui.nodes[self.id].selection_locked = self.window.controller.assistant.assistant_change_locked
        self.window.ui.nodes[self.id].setMinimumWidth(40)

        # rows
        layout = QVBoxLayout()
        layout.addWidget(header_widget)
        layout.addWidget(self.window.ui.nodes[self.id])

        # model
        self.window.ui.models[self.id] = self.create_model(self.window)
        self.window.ui.nodes[self.id].setModel(self.window.ui.models[self.id])

        return layout

    def create_model(self, parent):
        """
        Create list model

        :param parent: parent widget
        :return: QStandardItemModel
        :rtype: QStandardItemModel
        """
        return QStandardItemModel(0, 1, parent)

    def update(self, data):
        """
        Update list of assistants

        :param data: Data to update
        """
        # store previous selection
        self.window.ui.nodes[self.id].backup_selection()

        self.window.ui.models[self.id].removeRows(0, self.window.ui.models[self.id].rowCount())
        i = 0
        for n in data:
            self.window.ui.models[self.id].insertRow(i)
            name = data[n].name
            self.window.ui.models[self.id].setData(self.window.ui.models[self.id].index(i, 0), name)
            i += 1

        # restore previous selection
        self.window.ui.nodes[self.id].restore_selection()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

import json
import os
import shutil
from packaging.version import parse as parse_version

from pygpt_net.provider.preset.base import BaseProvider
from pygpt_net.item.preset import PresetItem


class JsonFileProvider(BaseProvider):
    def __init__(self, window=None):
        super(JsonFileProvider, self).__init__(window)
        self.window = window
        self.id = "json_file"
        self.type = "preset"
        self.config_dir = 'presets'

    def install(self):
        """
        Install provider data
        """
        # install presets
        presets_dir = os.path.join(self.window.app.config.path, self.config_dir)
        if not os.path.exists(presets_dir):
            src = os.path.join(self.window.app.config.get_root_path(), 'data', 'config', self.config_dir)
            shutil.copytree(src, presets_dir)
        else:
            # copy missing presets
            src = os.path.join(self.window.app.config.get_root_path(), 'data', 'config', self.config_dir)
            for file in os.listdir(src):
                src_file = os.path.join(src, file)
                dst_file = os.path.join(presets_dir, file)
                if not os.path.exists(dst_file):
                    shutil.copyfile(src_file, dst_file)

    def load(self):
        """
        Load presets from JSON file
        """
        items = {}
        path = os.path.join(self.window.app.config.path, self.config_dir)
        if not os.path.exists(path):
            print("FATAL ERROR: {} not found!".format(path))
            return None
        try:
            for filename in os.listdir(path):
                if filename.endswith(".json"):
                    path = os.path.join(self.window.app.config.path, self.config_dir, filename)
                    with open(path, 'r', encoding="utf-8") as f:
                        preset = PresetItem()
                        self.deserialize(json.load(f), preset)
                        items[filename[:-5]] = preset
        except Exception as e:
            self.window.app.debug.log(e)

        return items

    def save(self, id, item):
        """
        Save preset to JSON file

        :param item: PresetItem
        """
        path = os.path.join(self.window.app.config.path, self.config_dir, id + '.json')
        data = self.serialize(item)
        data['__meta__'] = self.window.app.config.append_meta()
        dump = json.dumps(data, indent=4)
        try:
            with open(path, 'w', encoding="utf-8") as f:
                f.write(dump)
        except Exception as e:
            self.window.app.debug.log(e)

    def save_all(self, items):
        """
        Save all presets to JSON files

        :param items: items dict
        """
        for id in items:
            path = os.path.join(self.window.app.config.path, self.config_dir, id + '.json')

            # serialize
            data = self.serialize(items[id])
            data['__meta__'] = self.window.app.config.append_meta()
            dump = json.dumps(data, indent=4)
            try:
                with open(path, 'w', encoding="utf-8") as f:
                    f.write(dump)
            except Exception as e:
                self.window.app.debug.log(e)

    def remove(self, id):
        """
        Remove preset

        :param id: preset id
        """
        path = os.path.join(self.window.app.config.path, self.config_dir, id + '.json')
        if os.path.exists(path):
            try:
                os.remove(path)
            except Exception as e:
                self.window.app.debug.log(e)

    def truncate(self):
        pass

    def patch(self, version):
        """
        Migrate presets to current app version

        :param version: current app version
        :return: true if migrated
        :rtype: bool
        """
        migrated = False
        for k in self.window.app.presets.items:
            data = self.window.app.presets.items[k]
            updated = False

            # get version of preset
            old = parse_version(data.version)

            # check if presets file is older than current app version
            if old < version:
                # < 2.0.0
                if old < parse_version("2.0.0"):
                    print("Migrating presets dir from < 2.0.0...")
                    self.window.app.updater.patch_file('presets', True)  # force replace file

                # < 2.0.53
                if old < parse_version("2.0.53") and k == 'current.assistant':
                    print("Migrating preset file from < 2.0.53...")
                    dst = os.path.join(self.window.app.config.path, 'presets', 'current.assistant.json')
                    src = os.path.join(self.window.app.config.get_root_path(), 'data', 'config', 'presets',
                                       'current.assistant.json')
                    shutil.copyfile(src, dst)
                    updated = True
                    print("Patched file: {}.".format(dst))

            # update file
            if updated:
                self.window.app.presets.load()  # reload presets from patched files
                self.window.app.presets.save(k)  # re-save presets
                migrated = True
                print("Preset {} patched to version {}.".format(k, version))

        return migrated

    @staticmethod
    def serialize(item):
        """
        Serialize item to dict

        :param item: item to serialize
        :return: serialized item
        :rtype: dict
        """
        return {
            'name': item.name,
            'ai_name': item.ai_name,
            'user_name': item.user_name,
            'prompt': item.prompt,
            'chat': item.chat,
            'completion': item.completion,
            'img': item.img,
            'vision': item.vision,
            'langchain': item.langchain,
            'assistant': item.assistant,
            'temperature': item.temperature,
            'filename': item.filename,
        }

    @staticmethod
    def deserialize(data, item):
        """
        Deserialize item from dict

        :param data: serialized item
        :param item: item to deserialize
        """
        if 'name' in data:
            item.name = data['name']
        if 'ai_name' in data:
            item.ai_name = data['ai_name']
        if 'user_name' in data:
            item.user_name = data['user_name']
        if 'prompt' in data:
            item.prompt = data['prompt']
        if 'chat' in data:
            item.chat = data['chat']
        if 'completion' in data:
            item.completion = data['completion']
        if 'img' in data:
            item.img = data['img']
        if 'vision' in data:
            item.vision = data['vision']
        if 'langchain' in data:
            item.langchain = data['langchain']
        if 'assistant' in data:
            item.assistant = data['assistant']
        if 'temperature' in data:
            item.temperature = data['temperature']
        if 'filename' in data:
            item.filename = data['filename']
        if '__meta__' in data and 'version' in data['__meta__']:
            item.version = data['__meta__']['version']

    def dump(self, item):
        """
        Dump to string

        :param item: item to dump
        :return: dumped item as string (json)
        :rtype: str
        """
        return json.dumps(self.serialize(item))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

from pygpt_net.provider.history.txt_file import TxtFileProvider


class History:

    def __init__(self, window):
        """
        History handler

        :param window: Window instance
        """
        self.window = window
        self.providers = {}
        self.provider = "txt_file"
        self.path = None

        # register data providers
        self.add_provider(TxtFileProvider())  # json file provider

    def add_provider(self, provider):
        """
        Add data provider

        :param provider: data provider instance
        """
        self.providers[provider.id] = provider
        self.providers[provider.id].window = self.window

    def install(self):
        """Install provider data"""
        if self.provider in self.providers:
            try:
                self.providers[self.provider].install()
            except Exception as e:
                self.window.app.debug.log(e)

    def append(self, ctx, mode):
        """
        Append text to history

        :param ctx: CtxItem instance
        :param mode: mode (input | output)
        """
        if self.provider in self.providers:
            try:
                self.providers[self.provider].append(ctx, mode)
            except Exception as e:
                if self.window is not None:
                    self.window.app.debug.log(e)
                else:
                    print("Error appending to history: {}".format(e))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #


class PresetItem:
    def __init__(self):
        self.name = "*"
        self.ai_name = ""
        self.user_name = ""
        self.prompt = ""
        self.chat = False
        self.completion = False
        self.img = False
        self.vision = False
        self.langchain = False
        self.assistant = False
        self.temperature = 1.0
        self.filename = None
        self.version = None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #


class ModelItem:
    def __init__(self, id=None):
        """
        Model data item

        :param id: Model ID
        """
        self.id = id
        self.name = None
        self.mode = []
        self.langchain = {}
        self.ctx = 0
        self.tokens = 0
        self.default = False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QVBoxLayout, QLabel, QHBoxLayout, QWidget, QCheckBox

from pygpt_net.utils import trans


class Vision:
    def __init__(self, window=None):
        """
        Toolbox UI

        :param window: Window instance
        """
        self.window = window

    def setup(self):
        """
        Setup vision

        :return: QWidget
        :rtype: QWidget
        """
        # enable/disable
        self.window.ui.nodes['vision.capture.enable'] = QCheckBox(trans("vision.capture.enable"))
        self.window.ui.nodes['vision.capture.enable'].stateChanged.connect(
            lambda: self.window.controller.camera.toggle(self.window.ui.nodes['vision.capture.enable'].isChecked()))
        self.window.ui.nodes['vision.capture.enable'].setToolTip(trans('vision.capture.enable.tooltip'))

        # auto
        self.window.ui.nodes['vision.capture.auto'] = QCheckBox(trans("vision.capture.auto"))
        self.window.ui.nodes['vision.capture.auto'].stateChanged.connect(
            lambda: self.window.controller.camera.toggle_auto(self.window.ui.nodes['vision.capture.auto'].isChecked()))
        self.window.ui.nodes['vision.capture.auto'].setToolTip(trans('vision.capture.auto.tooltip'))

        self.window.ui.nodes['vision.capture.label'] = QLabel(trans('vision.capture.options.title'))

        # checkbox options
        cols = QHBoxLayout()
        cols.addWidget(self.window.ui.nodes['vision.capture.enable'])
        cols.addWidget(self.window.ui.nodes['vision.capture.auto'])

        # rows
        rows = QVBoxLayout()
        rows.addWidget(self.window.ui.nodes['vision.capture.label'])
        rows.addLayout(cols)

        # widget
        self.window.ui.nodes['vision.capture.options'] = QWidget()
        self.window.ui.nodes['vision.capture.options'].setLayout(rows)
        self.window.ui.nodes['vision.capture.options'].setContentsMargins(0, 0, 0, 0)

        return self.window.ui.nodes['vision.capture.options']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

import json
import threading

from PySide6.QtCore import QObject, Signal, Slot


class Command:
    def __init__(self, window=None):
        """
        Command  controller

        :param window: Window instance
        """
        self.window = window
        self.thread = None
        self.thread_started = False
        self.force_stop = False

    def dispatch(self, event):
        """
        Dispatch cmd execute event (command execution)

        :param event: event object
        """
        self.dispatch_sync(event)

    def dispatch_sync(self, event):
        """
        Dispatch async event (command execution)

        :param event: event object
        """
        for id in self.window.app.plugins.plugins:
            if self.window.controller.plugins.is_enabled(id):
                if event.stop or self.window.controller.command.is_stop():
                    break
                self.window.app.dispatcher.apply(id, event, is_async=False)
        self.handle_finished(event)

    def dispatch_async(self, event):
        """
        Dispatch async event (command execution)

        :param event: event object
        """
        worker = CommandThread(window=self.window, event=event)
        worker.finished.connect(self.handle_finished)
        worker.destroyed.connect(self.handle_destroy)
        worker.debug.connect(self.handle_debug)

        self.thread = threading.Thread(target=worker.run)
        self.thread.start()
        self.thread_started = True

    def is_stop(self):
        """
        Check if stop is requested

        :return: true if stop is requested
        :rtype: bool
        """
        return self.force_stop

    @Slot(object)
    def handle_debug(self, data):
        """
        Handle thread debug log
        :param data
        """
        self.window.log(str(data))

    @Slot(object)
    def handle_finished(self, event):
        """
        Handle thread command execution finish
        :param event: event object
        """
        ctx = event.ctx
        self.window.set_status("")  # Clear status
        if ctx.reply:
            self.window.controller.input.send(json.dumps(ctx.results), force=True)
        self.thread_started = False

    @Slot()
    def handle_destroy(self):
        """Handle thread destroy"""
        self.thread_started = False


class CommandThread(QObject):
    finished = Signal(object)
    debug = Signal(object)
    destroyed = Signal()

    def __init__(self, window=None, event=None):
        """
        Run command dispatch thread

        :param window: Window instance
        :param event: event object
        """
        super().__init__()
        self.window = window
        self.event = event

    def run(self):
        """Run thread"""
        print("Starting command thread...")
        try:
            # dispatch event
            for id in self.window.app.plugins.plugins:
                if self.window.controller.plugins.is_enabled(id):
                    if self.event.stop or self.window.controller.command.is_stop():
                        break
                    self.window.app.dispatcher.apply(id, self.event, is_async=True)
            self.finished.emit(self.event)
        except Exception as e:
            print("Command thread error: " + str(e))
            self.window.app.debug.log(e)
        self.destroyed.emit()
        print("Command thread finished work.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

import datetime
import os

from pygpt_net.item.ctx import CtxItem, CtxMeta
from pygpt_net.core.tokens import num_tokens_from_context_item
from pygpt_net.provider.ctx.json_file import JsonFileProvider
from pygpt_net.utils import trans


class Ctx:
    def __init__(self, window=None):
        """
        Context handler

        :param window: Window instance
        """
        self.window = window
        self.providers = {}
        self.provider = "json_file"
        self.meta = {}
        self.items = []
        self.current = None
        self.assistant = None
        self.mode = None
        self.preset = None
        self.run = None
        self.status = None
        self.thread = None
        self.last_mode = None

        # register data providers
        self.add_provider(JsonFileProvider())  # json file provider

    def add_provider(self, provider):
        """
        Add ctx provider

        :param provider: ctx provider instance
        """
        self.providers[provider.id] = provider
        self.providers[provider.id].window = self.window

    def install(self):
        """Install provider data"""
        if self.provider in self.providers:
            try:
                self.providers[self.provider].install()
            except Exception as e:
                self.window.app.debug.log(e)

    def patch(self, app_version):
        """Patch provider data"""
        if self.provider in self.providers:
            try:
                self.providers[self.provider].patch(app_version)
            except Exception as e:
                self.window.app.debug.log(e)

    def select(self, id):
        """
        Select ctx

        :param id: context id
        """
        if id in self.meta:
            ctx = self.meta[id]
            self.current = id

            # reset
            self.thread = None
            self.mode = None
            self.assistant = None

            # restore
            self.thread = ctx.thread
            self.mode = ctx.mode
            self.assistant = ctx.assistant
            self.preset = ctx.preset

            self.items = self.load(id)

    def new(self):
        """
        Create new ctx and set as current

        :return: CtxMeta instance (new ctx meta)
        :rtype: CtxMeta
        """
        meta = self.create()  # create new ctx meta

        if meta is None:
            self.window.app.debug.log("Error creating new ctx")
            return

        self.meta[meta.id] = meta
        self.current = meta.id
        self.thread = None
        self.assistant = None
        self.mode = self.window.app.config.get('mode')
        self.preset = self.window.app.config.get('preset')
        self.items = []
        self.save(meta.id)

        return meta

    def build(self):
        """
        Build new ctx

        :return: created CtxMeta instance
        :rtype: CtxMeta
        """
        meta = CtxMeta()  # create ctx meta
        meta.name = "{}".format(trans('ctx.new.prefix'))
        meta.date = datetime.datetime.now().strftime("%Y-%m-%d")
        meta.mode = self.window.app.config.get('mode')
        meta.last_mode = self.window.app.config.get('mode')
        meta.initialized = False
        return meta

    def create(self):
        """
        Send created meta to provider and return new ID

        :return: created ID
        """
        meta = self.build()
        if self.provider in self.providers:
            try:
                id = self.providers[self.provider].create(meta)
                meta.id = id
                return meta
            except Exception as e:
                self.window.app.debug.log(e)

    def add(self, item):
        """
        Add CtxItem to contexts and saves context

        :param item: CtxItem to append
        """
        self.items.append(item)  # add CtxItem to context items

        # save context to file
        self.store()

    def is_empty(self):
        """
        Check if ctx is empty

        :return: true if empty, false otherwise
        :rtype: bool
        """
        if self.current is None:
            return True
        else:
            if len(self.items) == 0:
                return True
        return False

    def update(self):
        """
        Update current parent (when ctx item load from the list or setting mode)
        """
        self.mode = self.window.app.config.get('mode')

        if self.current is None:
            return
        self.meta[self.current].mode = self.mode
        self.save(self.current)

    def post_update(self, mode):
        """
        Update current (last) ctx data

        :param mode: mode name
        """
        if self.current is None:
            return

        # update current
        self.assistant = self.window.app.config.get('assistant')  # update assistant
        self.preset = self.window.app.config.get('preset')  # update preset

        # update current meta
        self.meta[self.current].last_mode = mode
        self.meta[self.current].preset = self.preset

        # if assistant then update assistant
        if mode == 'assistant':
            self.meta[self.current].assistant = self.assistant

        # save ctx
        self.save(self.current)

    def is_initialized(self):
        """
        Check if ctx is initialized (name assigned)

        :return: true if initialized, false otherwise
        :rtype: bool
        """
        if self.current is None:
            return
        if self.current in self.meta:
            return self.meta[self.current].initialized

    def set_initialized(self):
        """
        Set ctx as initialized (name assigned)
        """
        if self.current is None:
            return
        if self.current in self.meta:
            self.meta[self.current].initialized = True
            self.save(self.current)

    def has(self, id):
        """
        Check if ctx meta exists

        :param id: ctx ID
        :return: true if exists, false otherwise
        :rtype: bool
        """
        if id in self.meta:
            return True
        return False

    def get(self, idx):
        """
        Return ctx item by index

        :param idx: item index
        :return: context item
        :rtype: CtxItem or None
        """
        if idx < len(self.items):
            return self.items[idx]

    def get_meta(self):
        """
        Get ctx items sorted descending by date

        :return: ctx metas dict
        """
        return dict(sorted(self.meta.items(), reverse=True))

    def get_id_by_idx(self, idx):
        """
        Get ctx id (id) by index

        :param idx: index
        :return: ctx id
        :rtype: str or None
        """
        i = 0
        for id in self.get_meta():
            if i == idx:
                return id
            i += 1

    def get_idx_by_id(self, id):
        """
        Get ctx index by id

        :param id: id
        :return: idx
        :rtype: int or None
        """
        items = self.get_meta()
        if id in items:
            return list(items.keys()).index(id)

    def get_first(self):
        """
        Return first ctx ID from list

        :return: ctx id
        :rtype: str or None
        """
        for id in self.get_meta():
            return id

    def get_meta_by_id(self, id):
        """
        Return ctx meta by id

        :param id: ctx id
        :return: ctx dict
        :rtype: dict or None
        """
        if id in self.meta:
            return self.meta[id]

    def get_last(self):
        """
        Return last item from ctx

        :return: last ctx item
        :rtype: CtxItem or None
        """
        if len(self.items) > 0:
            return self.items[-1]
        return None

    def prepare(self):
        """Prepare context for prompt"""
        # if no contexts, create new one
        if len(self.meta) == 0:
            self.new()

    def count(self):
        """
        Count ctx items

        :return: ctx items count
        :rtype: int
        """
        return len(self.items)

    def all(self):
        """
        Return ctx items

        :return: ctx items
        :rtype: list
        """
        return self.items

    def remove(self, id):
        """
        Delete ctx by id

        :param id: ctx id
        """
        if id in self.meta:
            del self.meta[id]
            if self.provider in self.providers:
                try:
                    self.providers[self.provider].remove(id)
                except Exception as e:
                    self.window.app.debug.log(e)

    def truncate(self):
        """Delete all ctx"""
        # empty ctx index
        self.meta = {}

        # remove all ctx data in provider
        if self.provider in self.providers:
            try:
                self.providers[self.provider].truncate()
            except Exception as e:
                self.window.app.debug.log(e)

        # delete all txt history files from history dir
        path = os.path.join(self.window.app.config.path, 'history')
        for file in os.listdir(path):
            if file.endswith('.txt'):
                try:
                    os.remove(os.path.join(path, file))
                except Exception as e:
                    self.window.app.debug.log(e)

    def clear(self):
        """Clear ctx items"""
        self.items = []

    def append_thread(self, thread):
        """
        Append thread ID to ctx

        :param thread: thread ID
        """
        self.thread = thread
        if self.current is None:
            return
        if self.current in self.meta:
            self.meta[self.current].thread = self.thread
            self.save(self.current)

    def append_run(self, run):
        """
        Append run ID to ctx

        :param run: run ID
        """
        self.run = run
        if self.current is None:
            return
        if self.current in self.meta:
            self.meta[self.current].run = self.run
            self.save(self.current)

    def append_status(self, status):
        """
        Append status to ctx

        :param status: status
        """
        self.status = status
        if self.current is None:
            return
        if self.current in self.meta:
            self.meta[self.current].status = self.status
            self.save(self.current)

    def count_prompt_items(self, model, mode, used_tokens=100, max_tokens=1000):
        """
        Count ctx items to add to prompt

        :param model: model
        :param mode: mode
        :param used_tokens: used tokens
        :param max_tokens: max tokens
        :return: context items count, ctx tokens count
        :rtype: (int, int)
        """
        i = 0
        # loop on items from end to start
        tokens = used_tokens
        context_tokens = 0
        for item in reversed(self.items):
            num = num_tokens_from_context_item(item, mode, model)  # get num tokens for input and output
            tokens += num
            if tokens > max_tokens:
                break
            context_tokens += num
            i += 1

        return i, context_tokens

    def get_prompt_items(self, model, mode="chat", used_tokens=100, max_tokens=1000):
        """
        Return ctx items to add to prompt

        :param model: model
        :param mode: mode
        :param used_tokens: used tokens
        :param max_tokens: max tokens
        :return: context items list
        :rtype: list
        """
        items = []
        # loop on items from end to start
        tokens = used_tokens
        for item in reversed(self.items):
            tokens += num_tokens_from_context_item(item, mode, model)
            if tokens > max_tokens:
                break
            items.append(item)

        # reverse items
        items.reverse()
        return items

    def get_all_items(self):
        """
        Return all ctx items

        :return: ctx items list
        :rtype: list
        """
        items = []
        for item in reversed(self.items):
            items.append(item)

        # reverse items
        items.reverse()
        return items

    def check(self, threshold, max_total):
        """
        Check context and clear if limit exceeded

        :param threshold: threshold
        :param max_total: max total tokens
        """
        if self.get_tokens_left(max_total) <= threshold:
            self.remove_first()

    def get_tokens_left(self, max):
        """
        Return remaining tokens in context

        :param max: max tokens
        :return: remaining tokens in context
        :rtype: int
        """
        return max - self.get_total_tokens()

    def get_total_tokens(self):
        """
        Return current prompt tokens

        :return: total tokens
        :rtype: int
        """
        last = self.get_last()
        if last is not None:
            return last.total_tokens
        return 0

    def get_last_tokens(self):
        """
        Return last tokens count

        :return: last tokens
        :rtype: int
        """
        last = self.get_last()
        if last is not None:
            return last.total_tokens
        return 0

    def remove_last(self):
        """Remove last item"""
        if len(self.items) > 0:
            self.items.pop()

    def remove_first(self):
        """Remove first item"""
        if len(self.items) > 0:
            self.items.pop(0)

    def load_meta(self):
        """Load ctx list from provider"""
        if self.provider in self.providers:
            try:
                self.meta = self.providers[self.provider].get_meta()
            except Exception as e:
                self.window.app.debug.log(e)
                self.meta = {}

    def load(self, id):
        """
        Load ctx data from provider

        :param id: ctx id
        :return: ctx items
        :rtype: list
        """
        if self.provider in self.providers:
            try:
                return self.providers[self.provider].load(id)
            except Exception as e:
                self.window.app.debug.log(e)
        return []

    def save(self, id):
        """
        Save ctx data

        :param id: ctx id
        """
        if not self.window.app.config.get('store_history'):
            return

        if self.provider in self.providers:
            try:
                self.providers[self.provider].save(id, self.meta[id], self.items)
            except Exception as e:
                self.window.app.debug.log(e)

    def store(self):
        """Store current ctx"""
        if self.current is not None and self.current in self.meta:
            self.save(self.current)

    def dump(self, ctx):
        """
        Dump context item

        :param ctx: CtxItem instance
        """
        if self.provider in self.providers:
            try:
                return self.providers[self.provider].dump(ctx)
            except Exception as e:
                self.window.app.debug.log(e)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

class Dispatcher:
    def __init__(self, window=None):
        """
        Event dispatcher

        :param window: Window instance
        """
        self.window = window

    def dispatch(self, event, all=False, is_async=False):
        """
        Dispatch event to plugins

        :param event: event to dispatch
        :param all: true if dispatch to all plugins (enabled or not)
        :param is_async: true if async event
        """
        for id in self.window.app.plugins.plugins:
            if self.window.controller.plugins.is_enabled(id) or all:
                if event.stop:
                    break
                self.apply(id, event, is_async)

    def apply(self, id, event, is_async=False):
        """
        Handle event in plugin with provided id

        :param id: plugin id
        :param event: event object
        :param is_async: true if async event
        """
        if id in self.window.app.plugins.plugins:
            try:
                self.window.app.plugins.plugins[id].is_async = is_async
                self.window.app.plugins.plugins[id].handle(event)
            except AttributeError:
                pass


class Event:
    def __init__(self, name=None, data=None):
        """
        Event class

        :param name: event name
        :param data: event data
        """
        self.name = name
        self.data = data
        self.ctx = None
        self.stop = False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

import mimetypes
import os.path
import shutil
import ssl
from urllib.request import Request, urlopen

from pygpt_net.plugin.base_plugin import BasePlugin


class Plugin(BasePlugin):
    def __init__(self):
        super(Plugin, self).__init__()
        self.id = "cmd_files"
        self.name = "Command: Files I/O"
        self.description = "Provides commands to read and write files"
        self.window = None
        self.order = 100
        self.allowed_cmds = [
            "read_file",
            "save_file",
            "append_file",
            "delete_file",
            "list_dir",
            "mkdir",
            "download_file",
            "rmdir",
            "copy_file",
            "copy_dir",
            "move",
            "is_dir",
            "is_file",
            "file_exists",
            "file_size",
            "file_info",
        ]
        self.use_locale = True
        self.init_options()

    def init_options(self):
        """
        Initialize options
        """
        # cmd enable/disable
        self.add_option("cmd_read_file", "bool", True,
                        "Enable: Read file", "Allows `read_file` command execution")
        self.add_option("cmd_save_file", "bool", True,
                        "Enable: Save file", "Allows `save_file` command execution")
        self.add_option("cmd_append_file", "bool", True,
                        "Enable: Append to file", "Allows `append_file` command execution")
        self.add_option("cmd_delete_file", "bool", True,
                        "Enable: Delete file", "Allows `delete_file` command execution")
        self.add_option("cmd_list_dir", "bool", True,
                        "Enable: List files in directory (ls)", "Allows `list_dir` command execution")
        self.add_option("cmd_mkdir", "bool", True,
                        "Enable: Directory creation (mkdir)", "Allows `mkdir` command execution")
        self.add_option("cmd_download_file", "bool", True,
                        "Enable: Downloading files", "Allows `download_file` command execution")
        self.add_option("cmd_rmdir", "bool", True,
                        "Enable: Removing directories", "Allows `rmdir` command execution")
        self.add_option("cmd_copy_file", "bool", True,
                        "Enable: Copying files", "Allows `copy` command execution")
        self.add_option("cmd_copy_dir", "bool", True,
                        "Enable: Copying directories (recursive)", "Allows `copy_dir` command execution")
        self.add_option("cmd_move", "bool", True,
                        "Enable: Move files and directories (rename)", "Allows `move` command execution")
        self.add_option("cmd_is_dir", "bool", True,
                        "Enable: Check if path is directory", "Allows `is_dir` command execution")
        self.add_option("cmd_is_file", "bool", True,
                        "Enable: Check if path is file", "Allows `is_file` command execution")
        self.add_option("cmd_file_exists", "bool", True,
                        "Enable: Check if file or directory exists", "Allows `file_exists` command execution")
        self.add_option("cmd_file_size", "bool", True,
                        "Enable: Get file size", "Allows `file_size` command execution")
        self.add_option("cmd_file_info", "bool", True,
                        "Enable: Get file info", "Allows `file_info` command execution")

        # cmd syntax (prompt/instruction)
        self.add_option("syntax_read_file", "textarea", '"read_file": read data from file, params: "filename"',
                        "Syntax: read_file",
                        "Syntax for reading files", advanced=True)
        self.add_option("syntax_save_file", "textarea", '"save_file": save data to file, params: "filename", "data"',
                        "Syntax: save_file",
                        "Syntax for saving files", advanced=True)
        self.add_option("syntax_append_file", "textarea",
                        '"append_file": append data to file, params: "filename", "data"',
                        "Syntax: append_file",
                        "Syntax for appending to files", advanced=True)
        self.add_option("syntax_delete_file", "textarea", '"delete_file": delete file, params: "filename"',
                        "Syntax: delete_file",
                        "Syntax for deleting files", advanced=True)
        self.add_option("syntax_list_dir", "textarea", '"list_dir": list files and dirs in directory, params: "path"',
                        "Syntax: list_dir",
                        "Syntax for listing files", advanced=True)
        self.add_option("syntax_mkdir", "textarea", '"mkdir": create directory, params: "path"',
                        "Syntax: mkdir",
                        "Syntax for directory creation", advanced=True)
        self.add_option("syntax_download_file", "textarea", '"download_file": download file, params: "src", "dst"',
                        "Syntax: download_file",
                        "Syntax for downloading files", advanced=True)
        self.add_option("syntax_rmdir", "textarea", '"rmdir": remove directory, params: "path"',
                        "Syntax: rmdir",
                        "Syntax for removing directories", advanced=True)
        self.add_option("syntax_copy_file", "textarea", '"copy_file": copy file, params: "src", "dst"',
                        "Syntax: copy_file",
                        "Syntax for copying files", advanced=True)
        self.add_option("syntax_copy_dir", "textarea", '"copy_dir": recursive copy directory, params: "src", "dst"',
                        "Syntax: copy_dir",
                        "Syntax for recursive copying directories", advanced=True)
        self.add_option("syntax_move", "textarea", '"move": move file or directory, params: "src", "dst"',
                        "Syntax: move",
                        "Syntax for moving files and directories", advanced=True)
        self.add_option("syntax_is_dir", "textarea", '"is_dir": check if path is directory, params: "path"',
                        "Syntax: is_dir",
                        "Syntax for checking if path is directory", advanced=True)
        self.add_option("syntax_is_file", "textarea", '"is_file": check if path is file, params: "path"',
                        "Syntax: is_file",
                        "Syntax for checking if path is file", advanced=True)
        self.add_option("syntax_file_exists", "textarea", '"file_exists": check if file or directory exists, params: '
                                                          '"path"',
                        "Syntax: file_exists",
                        "Syntax for checking if file or directory exists", advanced=True)
        self.add_option("syntax_file_size", "textarea", '"file_size": get file size, params: "path"',
                        "Syntax: file_size",
                        "Syntax for getting file size", advanced=True)
        self.add_option("syntax_file_info", "textarea", '"file_info": get file info, params: "path"',
                        "Syntax: file_info",
                        "Syntax for getting file info", advanced=True)

    def setup(self):
        """
        Return available config options

        :return: config options
        :rtype: dict
        """
        return self.options

    def attach(self, window):
        """
        Attach window

        :param window: Window instance
        """
        self.window = window

    def handle(self, event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        """
        name = event.name
        data = event.data
        ctx = event.ctx

        if name == 'cmd.syntax':
            self.cmd_syntax(data)
        elif name == 'cmd.execute':
            self.cmd(ctx, data['commands'])

    def is_cmd_allowed(self, cmd):
        """
        Check if cmd is allowed

        :param cmd: command name
        :return: true if allowed
        :rtype: bool
        """
        key = "cmd_" + cmd
        if self.has_option(key) and self.get_option_value(key) is True:
            return True
        return False

    def log(self, msg):
        """
        Log message to console

        :param msg: message to log
        """
        self.debug('[CMD] ' + str(msg))
        print('[CMD] ' + str(msg))

    def cmd_syntax(self, data):
        """
        Event: On cmd syntax prepare

        :param data: event data dict
        """
        for option in self.allowed_cmds:
            if self.is_cmd_allowed(option):
                key = "syntax_" + option
                if self.has_option(key):
                    data['syntax'].append(str(self.get_option_value(key)))

    def cmd(self, ctx, cmds):
        """
        Event: On cmd

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        msg = None
        for item in cmds:
            try:
                if item["cmd"] in self.allowed_cmds and self.is_cmd_allowed(item["cmd"]):

                    # prepare request item for result
                    request_item = {"cmd": item["cmd"]}
                    ctx.reply = True  # send result message

                    # save file
                    if item["cmd"] == "save_file":
                        try:
                            msg = "Saving file: {}".format(item["params"]['filename'])
                            self.log(msg)
                            path = os.path.join(self.window.app.config.path, 'output', item["params"]['filename'])
                            data = item["params"]['data']
                            with open(path, 'w', encoding="utf-8") as file:
                                file.write(data)
                                file.close()
                                ctx.results.append({"request": request_item, "result": "OK"})
                                self.log("File saved: {}".format(path))
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error: {}".format(e)})
                            self.window.app.debug.log(e)
                            self.log("Error: {}".format(e))

                    # append to file
                    elif item["cmd"] == "append_file" and self.is_cmd_allowed("append_file"):
                        try:
                            msg = "Appending file: {}".format(item["params"]['filename'])
                            self.log(msg)
                            path = os.path.join(self.window.app.config.path, 'output', item["params"]['filename'])
                            data = item["params"]['data']
                            with open(path, 'a', encoding="utf-8") as file:
                                file.write(data)
                                file.close()
                                ctx.results.append({"request": request_item, "result": "OK"})
                                self.log("File appended: {}".format(path))
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error: {}".format(e)})
                            self.window.app.debug.log(e)
                            self.log("Error: {}".format(e))

                    # read file
                    elif item["cmd"] == "read_file" and self.is_cmd_allowed("read_file"):
                        try:
                            msg = "Reading file: {}".format(item["params"]['filename'])
                            self.log(msg)
                            path = os.path.join(self.window.app.config.path, 'output', item["params"]['filename'])
                            if os.path.exists(path):
                                with open(path, 'r', encoding="utf-8") as file:
                                    data = file.read()
                                    ctx.results.append({"request": request_item, "result": data})
                                    file.close()
                                    self.log("File read: {}".format(path))
                            else:
                                ctx.results.append({"request": request_item, "result": "File not found"})
                                self.log("File not found: {}".format(path))
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error: {}".format(e)})
                            self.window.app.debug.log(e)
                            self.log("Error: {}".format(e))

                    # delete file
                    elif item["cmd"] == "delete_file" and self.is_cmd_allowed("delete_file"):
                        try:
                            msg = "Deleting file: {}".format(item["params"]['filename'])
                            self.log(msg)
                            path = os.path.join(self.window.app.config.path, 'output', item["params"]['filename'])
                            if os.path.exists(path):
                                os.remove(path)
                                ctx.results.append({"request": request_item, "result": "OK"})
                                self.log("File deleted: {}".format(path))
                            else:
                                ctx.results.append({"request": request_item, "result": "File not found"})
                                self.log("File not found: {}".format(path))
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error: {}".format(e)})
                            self.window.app.debug.log(e)
                            self.log("Error: {}".format(e))

                    # list files
                    elif item["cmd"] == "list_dir" and self.is_cmd_allowed("list_dir"):
                        try:
                            msg = "Listing directory: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(self.window.app.config.path, 'output', item["params"]['path'])
                            if os.path.exists(path):
                                files = os.listdir(path)
                                ctx.results.append({"request": request_item, "result": files})
                                self.log("Files listed: {}".format(path))
                                self.log("Result: {}".format(files))
                            else:
                                ctx.results.append({"request": request_item, "result": "Directory not found"})
                                self.log("Directory not found: {}".format(path))
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error: {}".format(e)})
                            self.window.app.debug.log(e)
                            self.log("Error: {}".format(e))

                    # mkdir
                    elif item["cmd"] == "mkdir" and self.is_cmd_allowed("mkdir"):
                        try:
                            msg = "Creating directory: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(self.window.app.config.path, 'output', item["params"]['path'])
                            if not os.path.exists(path):
                                os.makedirs(path)
                                ctx.results.append({"request": request_item, "result": "OK"})
                                self.log("Directory created: {}".format(path))
                            else:
                                ctx.results.append({"request": request_item, "result": "Directory already exists"})
                                self.log("Directory already exists: {}".format(path))
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error: {}".format(e)})
                            self.window.app.debug.log(e)
                            self.log("Error: {}".format(e))

                    # rmdir
                    elif item["cmd"] == "rmdir" and self.is_cmd_allowed("rmdir"):
                        try:
                            msg = "Deleting directory: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(self.window.app.config.path, 'output', item["params"]['path'])
                            if os.path.exists(path):
                                shutil.rmtree(path)
                                ctx.results.append({"request": request_item, "result": "OK"})
                                self.log("Directory deleted: {}".format(path))
                            else:
                                ctx.results.append({"request": request_item, "result": "Directory not found"})
                                self.log("Directory not found: {}".format(path))
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error: {}".format(e)})
                            self.window.app.debug.log(e)
                            self.log("Error: {}".format(e))

                    # download
                    elif item["cmd"] == "download_file" and self.is_cmd_allowed("download_file"):
                        try:
                            dst = os.path.join(self.window.app.config.path, 'output', item["params"]['dst'])
                            msg = "Downloading file: {} into {}".format(item["params"]['src'], dst)
                            self.log(msg)

                            # Check if src is URL
                            if item["params"]['src'].startswith("http"):
                                src = item["params"]['src']
                                # Download file from URL
                                req = Request(
                                    url=src,
                                    headers={'User-Agent': 'Mozilla/5.0'}
                                )
                                context = ssl.create_default_context()
                                context.check_hostname = False
                                context.verify_mode = ssl.CERT_NONE
                                with urlopen(req, context=context, timeout=4) as response, open(dst, 'wb') as out_file:
                                    shutil.copyfileobj(response, out_file)
                            else:
                                # Handle local file paths
                                src = os.path.join(self.window.app.config.path, 'output', item["params"]['src'])

                                # Copy local file
                                with open(src, 'rb') as in_file, open(dst, 'wb') as out_file:
                                    shutil.copyfileobj(in_file, out_file)

                            # handle result
                            ctx.results.append({"request": request_item, "result": "OK"})
                            self.log("File downloaded: {} into {}".format(src, dst))
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error: {}".format(e)})
                            self.window.app.debug.log(e)
                            self.log("Error: {}".format(e))

                    # copy file
                    elif item["cmd"] == "copy_file" and self.is_cmd_allowed("copy_file"):
                        try:
                            msg = "Copying file: {} into {}".format(item["params"]['src'], item["params"]['dst'])
                            self.log(msg)
                            dst = os.path.join(self.window.app.config.path, 'output', item["params"]['dst'])
                            src = os.path.join(self.window.app.config.path, 'output', item["params"]['src'])
                            shutil.copyfile(src, dst)
                            ctx.results.append({"request": request_item, "result": "OK"})
                            self.log("File copied: {} into {}".format(src, dst))
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error: {}".format(e)})
                            self.window.app.debug.log(e)
                            self.log("Error: {}".format(e))

                    # copy dir
                    elif item["cmd"] == "copy_dir" and self.is_cmd_allowed("copy_dir"):
                        try:
                            msg = "Copying directory: {} into {}".format(item["params"]['src'], item["params"]['dst'])
                            self.log(msg)
                            dst = os.path.join(self.window.app.config.path, 'output', item["params"]['dst'])
                            src = os.path.join(self.window.app.config.path, 'output', item["params"]['src'])
                            shutil.copytree(src, dst)
                            ctx.results.append({"request": request_item, "result": "OK"})
                            self.log("Directory copied: {} into {}".format(src, dst))
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error {}".format(e)})
                            self.window.app.debug.log(e)
                            self.log("Error: {}".format(e))

                    # move
                    elif item["cmd"] == "move" and self.is_cmd_allowed("move"):
                        try:
                            msg = "Moving: {} into {}".format(item["params"]['src'], item["params"]['dst'])
                            self.log(msg)
                            dst = os.path.join(self.window.app.config.path, 'output', item["params"]['dst'])
                            src = os.path.join(self.window.app.config.path, 'output', item["params"]['src'])
                            shutil.move(src, dst)
                            ctx.results.append({"request": request_item, "result": "OK"})
                            self.log("Moved: {} into {}".format(src, dst))
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error: {}".format(e)})
                            self.window.app.debug.log(e)
                            self.log("Error: {}".format(e))

                    # is dir
                    elif item["cmd"] == "is_dir" and self.is_cmd_allowed("is_dir"):
                        try:
                            msg = "Checking if directory exists: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(self.window.app.config.path, 'output', item["params"]['path'])
                            if os.path.isdir(path):
                                ctx.results.append({"request": request_item, "result": "OK"})
                                self.log("Directory exists: {}".format(path))
                            else:
                                ctx.results.append({"request": request_item, "result": "Directory not found"})
                                self.log("Directory not found: {}".format(path))
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error: {}".format(e)})
                            self.window.app.debug.log(e)
                            self.log("Error: {}".format(e))

                    # is file
                    elif item["cmd"] == "is_file" and self.is_cmd_allowed("is_file"):
                        try:
                            msg = "Checking if file exists: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(self.window.app.config.path, 'output', item["params"]['path'])
                            if os.path.isfile(path):
                                ctx.results.append({"request": request_item, "result": "OK"})
                                self.log("File exists: {}".format(path))
                            else:
                                ctx.results.append({"request": request_item, "result": "File not found"})
                                self.log("File not found: {}".format(path))
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error: {}".format(e)})
                            self.window.app.debug.log(e)
                            self.log("Error: {}".format(e))

                    # file exists
                    elif item["cmd"] == "file_exists" and self.is_cmd_allowed("file_exists"):
                        try:
                            msg = "Checking if path exists: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(self.window.app.config.path, 'output', item["params"]['path'])
                            if os.path.exists(path):
                                ctx.results.append({"request": request_item, "result": "OK"})
                                self.log("Path exists: {}".format(path))
                            else:
                                ctx.results.append({"request": request_item, "result": "File or directory not found"})
                                self.log("Path not found: {}".format(path))
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error: {}".format(e)})
                            self.window.app.debug.log(e)
                            self.log("Error: {}".format(e))

                    # file size
                    elif item["cmd"] == "file_size" and self.is_cmd_allowed("file_size"):
                        try:
                            msg = "Checking file size: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(self.window.app.config.path, 'output', item["params"]['path'])
                            if os.path.exists(path):
                                ctx.results.append({"request": request_item, "result": os.path.getsize(path)})
                                self.log("File size: {}".format(os.path.getsize(path)))
                            else:
                                ctx.results.append({"request": request_item, "result": "File not found"})
                                self.log("File not found: {}".format(path))
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error: {}".format(e)})
                            self.window.app.debug.log(e)
                            self.log("Error: {}".format(e))

                    # file info
                    elif item["cmd"] == "file_info" and self.is_cmd_allowed("file_info"):
                        try:
                            msg = "Checking file info: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(self.window.app.config.path, 'output', item["params"]['path'])
                            if os.path.exists(path):
                                data = {
                                    "size": os.path.getsize(path),
                                    'mime_type': mimetypes.guess_type(path)[0] or 'application/octet-stream',
                                    "last_access": os.path.getatime(path),
                                    "last_modification": os.path.getmtime(path),
                                    "creation_time": os.path.getctime(path),
                                    "is_dir": os.path.isdir(path),
                                    "is_file": os.path.isfile(path),
                                    "is_link": os.path.islink(path),
                                    "is_mount": os.path.ismount(path),
                                    'stat': os.stat(path),
                                }
                                ctx.results.append({"request": request_item, "result": data})
                                self.log("File info: {}".format(data))
                            else:
                                ctx.results.append({"request": request_item, "result": "File not found"})
                                self.log("File not found: {}".format(path))
                        except Exception as e:
                            ctx.results.append({"request": request_item, "result": "Error: {}".format(e)})
                            self.window.app.debug.log(e)
                            self.log("Error: {}".format(e))
            except Exception as e:
                ctx.results.append({"request": item, "result": "Error: {}".format(e)})
                ctx.reply = True
                self.window.app.debug.log(e)
                self.log("Error: {}".format(e))

        if msg is not None:
            self.window.set_status(msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

class ContextDebug:
    def __init__(self, window=None):
        """
        Context debug

        :param window: Window instance
        """
        self.window = window
        self.id = 'context'

    def update(self):
        """Update debug window"""
        self.window.app.debug.begin(self.id)
        self.window.app.debug.add(self.id, 'current', str(self.window.app.ctx.current))
        self.window.app.debug.add(self.id, 'assistant', str(self.window.app.ctx.assistant))
        self.window.app.debug.add(self.id, 'mode', str(self.window.app.ctx.mode))
        self.window.app.debug.add(self.id, 'preset', str(self.window.app.ctx.preset))
        self.window.app.debug.add(self.id, 'run', str(self.window.app.ctx.run))
        self.window.app.debug.add(self.id, 'status', str(self.window.app.ctx.status))
        self.window.app.debug.add(self.id, 'thread', str(self.window.app.ctx.thread))

        current = None
        if self.window.app.ctx.current is not None:
            if self.window.app.ctx.current in self.window.app.ctx.meta:
                current = self.window.app.ctx.meta[self.window.app.ctx.current]
            if current is not None:
                self.window.app.debug.add(self.id, '[current] id', str(current.id))
                self.window.app.debug.add(self.id, '[current] assistant', str(current.assistant))
                self.window.app.debug.add(self.id, '[current] date', str(current.date))
                self.window.app.debug.add(self.id, '[current] last_mode', str(current.last_mode))
                self.window.app.debug.add(self.id, '[current] mode', str(current.mode))
                self.window.app.debug.add(self.id, '[current] name', str(current.name))
                self.window.app.debug.add(self.id, '[current] preset', str(current.preset))
                self.window.app.debug.add(self.id, '[current] run', str(current.run))
                self.window.app.debug.add(self.id, '[current] status', str(current.status))
                self.window.app.debug.add(self.id, '[current] thread', str(current.thread))

        self.window.app.debug.add(self.id, 'len(contexts)', str(len(self.window.app.ctx.meta)))
        self.window.app.debug.add(self.id, 'len(items)', str(len(self.window.app.ctx.items)))

        i = 0
        for item in self.window.app.ctx.items:
            prefix = '[{}] '.format(i)
            self.window.app.debug.add(self.id, prefix + 'mode', str(item.mode))
            self.window.app.debug.add(self.id, prefix + 'input_name', str(item.input_name))
            self.window.app.debug.add(self.id, prefix + 'input', str(item.input))
            self.window.app.debug.add(self.id, prefix + 'output_name', str(item.output_name))
            self.window.app.debug.add(self.id, prefix + 'output', str(item.output))
            self.window.app.debug.add(self.id, prefix + 'input_tokens',
                                      str(item.input_tokens))
            self.window.app.debug.add(self.id, prefix + 'output_tokens',
                                      str(item.output_tokens))
            self.window.app.debug.add(self.id, prefix + 'total_tokens', str(item.total_tokens))
            self.window.app.debug.add(self.id, '------', '')
            i += 1

        self.window.app.debug.end(self.id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

import os


class Filesystem:
    def __init__(self, window=None):
        """
        Filesystem handler

        :param window: Window instance
        """
        self.window = window

    def install(self):
        """Install provider data"""
        img_dir = os.path.join(self.window.app.config.path, 'output')
        if not os.path.exists(img_dir):
            os.mkdir(img_dir)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

import os

import cv2

from PySide6.QtCore import QObject, Signal


class Camera:
    def __init__(self, window=None):
        """
        Camera handler

        :param window: Window instance
        """
        self.window = window
        self.capture = None
        self.current = None

    def install(self):
        """Install provider data"""
        img_dir = os.path.join(self.window.app.config.path, 'capture')
        if not os.path.exists(img_dir):
            os.mkdir(img_dir)


class CameraThread(QObject):
    finished = Signal()
    destroyed = Signal()
    started = Signal()
    stopped = Signal()

    def __init__(self, window=None):
        """
        Camera capture thread

        :param window: Window instance
        """
        super().__init__()
        self.window = window
        self.initialized = False
        self.capture = None
        self.frame = None

    def setup_camera(self):
        """
        Initialize camera
        """
        try:
            # get params from global config
            self.capture = cv2.VideoCapture(self.window.app.config.get('vision.capture.idx'))
            self.capture.set(cv2.CAP_PROP_FRAME_WIDTH, self.window.app.config.get('vision.capture.width'))
            self.capture.set(cv2.CAP_PROP_FRAME_HEIGHT, self.window.app.config.get('vision.capture.height'))
        except Exception as e:
            self.window.app.debug.log(e)
            print("Camera thread setup exception", e)
            self.finished.emit(e)

    def run(self):
        """
        Frame capture loop
        """
        try:
            if not self.initialized:
                self.setup_camera()
                self.started.emit()
                self.initialized = True

            while True:
                if self.window.is_closing \
                        or self.capture is None \
                        or not self.capture.isOpened() \
                        or self.window.controller.camera.stop:
                    self.release()  # release camera
                    self.stopped.emit()
                    break
                _, frame = self.capture.read()
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                frame = cv2.flip(frame, 1)
                self.window.controller.camera.frame = frame  # update frame
        except Exception as e:
            self.window.app.debug.log(e)
            print("Camera thread capture exception", e)

        # release camera
        self.release()
        self.finished.emit()

    def release(self):
        """
        Release camera
        """
        if self.capture is not None and self.capture.isOpened():
            self.capture.release()
        self.capture = None
        self.frame = None
        self.initialized = False
